from chalk._gen.chalk.chart.v1 import densetimeserieschart_pb2 as _densetimeserieschart_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureValueBaseWindowFunction(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_UNSPECIFIED: _ClassVar[FeatureValueBaseWindowFunction]
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_UNIQUE_VALUES: _ClassVar[FeatureValueBaseWindowFunction]
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_TOTAL_OBSERVATIONS: _ClassVar[FeatureValueBaseWindowFunction]
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_NULL_PERCENTAGE: _ClassVar[FeatureValueBaseWindowFunction]
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_MAX_VALUE: _ClassVar[FeatureValueBaseWindowFunction]
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_MIN_VALUE: _ClassVar[FeatureValueBaseWindowFunction]
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_AVERAGE: _ClassVar[FeatureValueBaseWindowFunction]
    FEATURE_VALUE_BASE_WINDOW_FUNCTION_UNIQUE_PKEYS: _ClassVar[FeatureValueBaseWindowFunction]

class FeatureValueGroupBy(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FEATURE_VALUE_GROUP_BY_UNSPECIFIED: _ClassVar[FeatureValueGroupBy]
    FEATURE_VALUE_GROUP_BY_CATEGORICAL_VALUE: _ClassVar[FeatureValueGroupBy]
    FEATURE_VALUE_GROUP_BY_PRIMARY_KEY: _ClassVar[FeatureValueGroupBy]
    FEATURE_VALUE_GROUP_BY_DEPLOYMENT_ID: _ClassVar[FeatureValueGroupBy]
    FEATURE_VALUE_GROUP_BY_RESOLVER: _ClassVar[FeatureValueGroupBy]
    FEATURE_VALUE_GROUP_BY_OPERATION_KIND: _ClassVar[FeatureValueGroupBy]
    FEATURE_VALUE_GROUP_BY_OPERATION_ID: _ClassVar[FeatureValueGroupBy]
    FEATURE_VALUE_GROUP_BY_FEATURE_VERSION: _ClassVar[FeatureValueGroupBy]

class FeatureValueTimeSeries(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FEATURE_VALUE_TIME_SERIES_UNSPECIFIED: _ClassVar[FeatureValueTimeSeries]
    FEATURE_VALUE_TIME_SERIES_INSERTED_AT: _ClassVar[FeatureValueTimeSeries]
    FEATURE_VALUE_TIME_SERIES_OBSERVED_AT: _ClassVar[FeatureValueTimeSeries]

FEATURE_VALUE_BASE_WINDOW_FUNCTION_UNSPECIFIED: FeatureValueBaseWindowFunction
FEATURE_VALUE_BASE_WINDOW_FUNCTION_UNIQUE_VALUES: FeatureValueBaseWindowFunction
FEATURE_VALUE_BASE_WINDOW_FUNCTION_TOTAL_OBSERVATIONS: FeatureValueBaseWindowFunction
FEATURE_VALUE_BASE_WINDOW_FUNCTION_NULL_PERCENTAGE: FeatureValueBaseWindowFunction
FEATURE_VALUE_BASE_WINDOW_FUNCTION_MAX_VALUE: FeatureValueBaseWindowFunction
FEATURE_VALUE_BASE_WINDOW_FUNCTION_MIN_VALUE: FeatureValueBaseWindowFunction
FEATURE_VALUE_BASE_WINDOW_FUNCTION_AVERAGE: FeatureValueBaseWindowFunction
FEATURE_VALUE_BASE_WINDOW_FUNCTION_UNIQUE_PKEYS: FeatureValueBaseWindowFunction
FEATURE_VALUE_GROUP_BY_UNSPECIFIED: FeatureValueGroupBy
FEATURE_VALUE_GROUP_BY_CATEGORICAL_VALUE: FeatureValueGroupBy
FEATURE_VALUE_GROUP_BY_PRIMARY_KEY: FeatureValueGroupBy
FEATURE_VALUE_GROUP_BY_DEPLOYMENT_ID: FeatureValueGroupBy
FEATURE_VALUE_GROUP_BY_RESOLVER: FeatureValueGroupBy
FEATURE_VALUE_GROUP_BY_OPERATION_KIND: FeatureValueGroupBy
FEATURE_VALUE_GROUP_BY_OPERATION_ID: FeatureValueGroupBy
FEATURE_VALUE_GROUP_BY_FEATURE_VERSION: FeatureValueGroupBy
FEATURE_VALUE_TIME_SERIES_UNSPECIFIED: FeatureValueTimeSeries
FEATURE_VALUE_TIME_SERIES_INSERTED_AT: FeatureValueTimeSeries
FEATURE_VALUE_TIME_SERIES_OBSERVED_AT: FeatureValueTimeSeries

class FeatureValuePercentileWindowFunction(_message.Message):
    __slots__ = ("percentile",)
    PERCENTILE_FIELD_NUMBER: _ClassVar[int]
    percentile: float
    def __init__(self, percentile: _Optional[float] = ...) -> None: ...

class FeatureValueSeries(_message.Message):
    __slots__ = ("feature_fqn", "series_title", "base_window_function", "percentile_window_function")
    FEATURE_FQN_FIELD_NUMBER: _ClassVar[int]
    SERIES_TITLE_FIELD_NUMBER: _ClassVar[int]
    BASE_WINDOW_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    PERCENTILE_WINDOW_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    feature_fqn: str
    series_title: str
    base_window_function: FeatureValueBaseWindowFunction
    percentile_window_function: FeatureValuePercentileWindowFunction
    def __init__(
        self,
        feature_fqn: _Optional[str] = ...,
        series_title: _Optional[str] = ...,
        base_window_function: _Optional[_Union[FeatureValueBaseWindowFunction, str]] = ...,
        percentile_window_function: _Optional[_Union[FeatureValuePercentileWindowFunction, _Mapping]] = ...,
    ) -> None: ...

class GetFeatureValuesTimeSeriesChartRequest(_message.Message):
    __slots__ = (
        "title",
        "series",
        "window_period",
        "start_timestamp_inclusive",
        "end_timestamp_exclusive",
        "time_series",
        "group_by",
        "group_by_series_limit",
    )
    TITLE_FIELD_NUMBER: _ClassVar[int]
    SERIES_FIELD_NUMBER: _ClassVar[int]
    WINDOW_PERIOD_FIELD_NUMBER: _ClassVar[int]
    START_TIMESTAMP_INCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    END_TIMESTAMP_EXCLUSIVE_FIELD_NUMBER: _ClassVar[int]
    TIME_SERIES_FIELD_NUMBER: _ClassVar[int]
    GROUP_BY_FIELD_NUMBER: _ClassVar[int]
    GROUP_BY_SERIES_LIMIT_FIELD_NUMBER: _ClassVar[int]
    title: str
    series: _containers.RepeatedCompositeFieldContainer[FeatureValueSeries]
    window_period: _duration_pb2.Duration
    start_timestamp_inclusive: _timestamp_pb2.Timestamp
    end_timestamp_exclusive: _timestamp_pb2.Timestamp
    time_series: FeatureValueTimeSeries
    group_by: _containers.RepeatedScalarFieldContainer[FeatureValueGroupBy]
    group_by_series_limit: int
    def __init__(
        self,
        title: _Optional[str] = ...,
        series: _Optional[_Iterable[_Union[FeatureValueSeries, _Mapping]]] = ...,
        window_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        start_timestamp_inclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        end_timestamp_exclusive: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        time_series: _Optional[_Union[FeatureValueTimeSeries, str]] = ...,
        group_by: _Optional[_Iterable[_Union[FeatureValueGroupBy, str]]] = ...,
        group_by_series_limit: _Optional[int] = ...,
    ) -> None: ...

class GetFeatureValuesTimeSeriesChartResponse(_message.Message):
    __slots__ = ("chart",)
    CHART_FIELD_NUMBER: _ClassVar[int]
    chart: _densetimeserieschart_pb2.DenseTimeSeriesChart
    def __init__(
        self, chart: _Optional[_Union[_densetimeserieschart_pb2.DenseTimeSeriesChart, _Mapping]] = ...
    ) -> None: ...



from __future__ import annotations

import contextlib
import io
import logging
import os
import threading
import typing
import uuid
from concurrent.futures import Future, ThreadPoolExecutor
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Mapping, Optional, Sequence, Type, Union

import pyarrow as pa
import pyarrow.parquet as pq

import chalk.logging
from chalk.features import Feature, FeatureConverter
from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.integrations.util import json_parse_and_cast
from chalk.sql._internal.query_execution_parameters import QueryExecutionParameters
from chalk.sql._internal.sql_source import (
    BaseSQLSource,
    SQLSourceKind,
    UnsupportedEfficientExecutionError,
    validate_dtypes_for_efficient_execution,
)
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.utils.df_utils import is_binary_like, read_parquet
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.threading import DEFAULT_IO_EXECUTOR

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3Client
    from sqlalchemy.engine import Connection
    from sqlalchemy.engine.url import URL
    from sqlalchemy.types import TypeEngine

_logger = get_logger(__name__)
_public_logger = chalk.logging.chalk_logger


def get_supported_redshift_unload_types() -> List[Type["TypeEngine"]]:
    """
    This is a method instead of a constant to avoid import issues
    :return:
    """

    import sqlalchemy as sa

    return [
        # SQL Standard types
        sa.JSON,
        sa.ARRAY,
        sa.REAL,
        sa.FLOAT,
        sa.NUMERIC,
        sa.DECIMAL,
        sa.INTEGER,
        sa.SMALLINT,
        sa.BIGINT,
        sa.TIMESTAMP,
        sa.DATETIME,
        sa.DATE,
        sa.TIME,
        sa.TEXT,
        sa.CLOB,
        sa.VARCHAR,
        sa.NVARCHAR,
        sa.CHAR,
        sa.NCHAR,
        sa.BLOB,
        sa.BINARY,
        # sa.VARBINARY, # This dtype is not supported for UNLOAD into PARQUET
        sa.BOOLEAN,
        # Generic types
        sa.Integer,
        sa.Float,
        sa.Numeric,
        sa.SmallInteger,
        sa.BigInteger,
        sa.DateTime,
        sa.Date,
        sa.Time,
        sa.Text,
        sa.Boolean,
    ]


_DEFAULT_REDSHIFT_S3_CLIENT = None
_BOTO_CLIENT_LOCK = threading.Lock()
_BOTO_SERIALIZE_CALLS: bool = env_var_bool("CHALK_REDSHIFT_SERIALIZE_BOTO_CALLS")


_REDSHIFT_HOST_NAME = "REDSHIFT_HOST"
_REDSHIFT_DB_NAME = "REDSHIFT_DB"
_REDSHIFT_USER_NAME = "REDSHIFT_USER"
_REDSHIFT_PASSWORD_NAME = "REDSHIFT_PASSWORD"
_REDSHIFT_UNLOAD_IAM_ROLE_NAME = "REDSHIFT_UNLOAD_IAM_ROLE"
_REDSHIFT_PORT_NAME = "REDSHIFT_PORT"
_REDSHIFT_UNLOAD_S3_BUCKET_NAME = "REDSHIFT_UNLOAD_S3_BUCKET"


class RedshiftSourceImpl(BaseSQLSource):
    kind = SQLSourceKind.redshift

    def __init__(
        self,
        host: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
        unload_iam_role: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
        s3_client: Optional[S3Client] = None,
        s3_bucket: Optional[str] = None,
        executor: Optional[ThreadPoolExecutor] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import boto3
            import redshift_connector
        except ImportError:
            raise missing_dependency_exception("chalkpy[redshift]")
        del redshift_connector
        self.host = host or load_integration_variable(
            name=_REDSHIFT_HOST_NAME, integration_name=name, override=integration_variable_override
        )
        self.db = db or load_integration_variable(
            name=_REDSHIFT_DB_NAME, integration_name=name, override=integration_variable_override
        )
        self.user = user or load_integration_variable(
            name=_REDSHIFT_USER_NAME, integration_name=name, override=integration_variable_override
        )
        self.password = password or load_integration_variable(
            name=_REDSHIFT_PASSWORD_NAME, integration_name=name, override=integration_variable_override
        )
        self.unload_iam_role = unload_iam_role or load_integration_variable(
            name=_REDSHIFT_UNLOAD_IAM_ROLE_NAME, integration_name=name, override=integration_variable_override
        )
        self.port = (
            int(port)
            if port is not None
            else load_integration_variable(
                name=_REDSHIFT_PORT_NAME, integration_name=name, parser=int, override=integration_variable_override
            )
        )
        # TODO customer may want to provide the s3 creds via the "AWS Integration" we provide.
        # NOTE: Chalk Engine injects its own S3 client via `_DEFAULT_REDSHIFT_S3_CLIENT`
        self._s3_client = s3_client or _DEFAULT_REDSHIFT_S3_CLIENT or boto3.client("s3")
        self._s3_bucket = s3_bucket or load_integration_variable(
            name=_REDSHIFT_UNLOAD_S3_BUCKET_NAME, integration_name=name, override=integration_variable_override
        )
        if self._s3_bucket is not None:
            self._s3_bucket = self._s3_bucket.removesuffix("s3://").lstrip("/")
        self._executor = executor or DEFAULT_IO_EXECUTOR

        if engine_args is None:
            engine_args = {}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        engine_args.setdefault(
            "connect_args",
            {
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )
        # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
        # transactions are not needed
        # Setting the isolation level on the engine, instead of the connection, avoids
        # a DBAPI statement to reset the transactional level back to the default before returning the
        # connection to the pool
        engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args={})

    @property
    def s3_bucket(self):
        return self._s3_bucket

    def convert_db_types(self, v: Any, converter: FeatureConverter):
        """
        Redshift returns binary data as a hex string. Need to convert to bytes before returning as feature data.
        """
        if is_binary_like(converter.pyarrow_dtype):
            assert isinstance(v, str)
            v = bytes.fromhex(v)
        return super().convert_db_types(v, converter)

    def get_sqlglot_dialect(self) -> str | None:
        return "redshift"

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="redshift+psycopg2",
            username=self.user,
            password=self.password,
            host=self.host,
            database=self.db,
            port=self.port,
        )

    def supports_inefficient_fallback(self) -> bool:
        return False

    def _execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> Iterable[pa.RecordBatch]:
        temp_query_id = id(finalized_query)
        _public_logger.debug(f"Executing RedShift query [{temp_query_id}]...")

        if self._s3_bucket is None:
            raise UnsupportedEfficientExecutionError(
                "No S3 destination bucket configured for Redshift UNLOAD command, falling back to default execution mode.",
                log_level=logging.INFO,
            )
        with self.get_engine().connect() if connection is None else contextlib.nullcontext(connection) as cnx:
            with cnx.begin():
                cursor = cnx.connection.cursor()
                compiled_statement = self._get_compiled_query(finalized_query)

                from sqlalchemy.sql import Select

                if isinstance(finalized_query.query, Select):
                    validate_dtypes_for_efficient_execution(
                        finalized_query.query, supported_types=get_supported_redshift_unload_types()
                    )
                execution_context = self.get_engine().dialect.execution_ctx_cls._init_compiled(  # type: ignore
                    connection=cnx,
                    dialect=cnx.dialect,
                    dbapi_connection=cnx.connection.dbapi_connection,
                    execution_options={},
                    compiled=compiled_statement,
                    parameters=[],
                    invoked_statement=None,
                    extracted_parameters=None,
                )
                assert isinstance(execution_context, self.get_engine().dialect.execution_ctx_cls)
                operation = execution_context.statement
                assert operation is not None
                operation = operation.strip().removesuffix(";")
                params = execution_context.parameters[0]
                unload_destination = None

                # Redshift requires that the query used for an unload statement is literal quoted. That means that we CANNOT use a bind parameters within it
                # So, we will get around this using a "CREATE TEMP TABLE as SELECT (query with bind parameters)"
                # Then, we will run an unload query without any parameters: UNLOAD ('select * from temp_table')
                # Finally, we can drop the temp table
                temp_table_name = f"query_{str(uuid.uuid4()).replace('-', '_')}"
                try:
                    _logger.debug(f"Executing query & creating temp table '{temp_table_name}'")
                    cursor.execute(f"CREATE TEMP TABLE {temp_table_name} AS ({operation})", params)
                except Exception as e:
                    _public_logger.error(f"Failed to create temp table for operation: {operation}", exc_info=e)
                    raise RuntimeError(f"Failed to create temp table for operation: {operation}") from e
                else:
                    unload_destination = f"{temp_table_name}/"
                    unload_uri = f"s3://{self._s3_bucket}/{unload_destination}"
                    unload_iam_role = "default" if self.unload_iam_role is None else f"'{self.unload_iam_role}'"
                    unload_query = f"UNLOAD ('SELECT * FROM {temp_table_name}') TO '{unload_uri}' IAM_ROLE {unload_iam_role} FORMAT PARQUET EXTENSION 'parquet'"
                    _public_logger.info(f"Executing UNLOAD query [{temp_query_id}]: {unload_query}")
                    cursor.execute(unload_query)
                finally:
                    try:
                        cursor.execute(f"DROP TABLE {temp_table_name}")
                    except Exception:
                        _logger.warning(f"Failed to drop temp table '{temp_table_name}'", exc_info=True)
        # Redshift is case-insensitive with column names, so let's map it back to what we were expecting
        assert unload_destination is not None
        _public_logger.info(
            f"Finished executing redshift UNLOAD, [{temp_query_id}] reading parquet data from {unload_destination}..."
        )
        yield from self._download_objs_async(
            unload_destination=unload_destination,
            columns_to_features=columns_to_features,
            yield_empty_batches=query_execution_parameters.yield_empty_batches,
        )

    def _download_objs_async(
        self,
        unload_destination: str,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        yield_empty_batches: bool,
    ) -> Iterable[pa.RecordBatch]:
        assert self._s3_bucket is not None
        filenames = list(_list_files(self._s3_client, self._s3_bucket, unload_destination))
        download_futs: list[Future[pa.Table]] = []
        for filename in filenames:
            _logger.debug("Scheduling download of file %s", filename)
            fut = self._executor.submit(
                _download_file_to_table, self._s3_client, self._s3_bucket, filename, columns_to_features
            )
            download_futs.append(fut)
        _public_logger.info(f"Downloading parquet data partitioned into {len(download_futs)} files...")
        schema: pa.Schema | None = None
        yielded = False
        for fut in download_futs:
            tbl = fut.result()
            if len(tbl) > 0:
                yield tbl.combine_chunks().to_batches()[0]
                yielded = True
            if len(tbl) == 0 and schema is None:
                schema = tbl.schema
        if not yielded and schema is not None:
            yield pa.RecordBatch.from_pydict({k: [] for k in schema.names}, schema)

    @classmethod
    def register_sqlalchemy_compiler_overrides(cls):
        try:
            from chalk.sql._internal.integrations.redshift_compiler_overrides import register_redshift_compiler_hooks
        except ImportError:
            raise missing_dependency_exception("chalkpy[redshift]")

        register_redshift_compiler_hooks()

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_REDSHIFT_HOST_NAME, self.name, self.host),
                create_integration_variable(_REDSHIFT_DB_NAME, self.name, self.db),
                create_integration_variable(_REDSHIFT_USER_NAME, self.name, self.user),
                create_integration_variable(_REDSHIFT_PASSWORD_NAME, self.name, self.password),
                create_integration_variable(_REDSHIFT_PORT_NAME, self.name, self.port),
                create_integration_variable(_REDSHIFT_UNLOAD_IAM_ROLE_NAME, self.name, self.unload_iam_role),
                create_integration_variable(_REDSHIFT_UNLOAD_S3_BUCKET_NAME, self.name, self._s3_bucket),
            ]
            if v is not None
        }


def _boto_lock_ctx() -> typing.ContextManager:
    if _BOTO_SERIALIZE_CALLS:
        lock_ctx = _BOTO_CLIENT_LOCK
    else:
        lock_ctx = contextlib.nullcontext()
    return lock_ctx


def _list_files(client: S3Client, bucket: str, prefix: str) -> Iterable[str]:
    try:
        continuation_token = None
        while True:
            with _boto_lock_ctx():
                if continuation_token is None:
                    resp = client.list_objects_v2(
                        Bucket=bucket,
                        Prefix=prefix,
                    )
                else:
                    resp = client.list_objects_v2(
                        Bucket=bucket,
                        Prefix=prefix,
                        ContinuationToken=continuation_token,
                    )
            # If no keys returned, the server omits 'Contents'
            for row in resp.get("Contents", []):
                key = row.get("Key")
                assert key is not None, "all objects must have a key"
                yield key
            if not resp["IsTruncated"]:
                return
            continuation_token = resp["NextContinuationToken"]
    except Exception:
        _logger.error(f"Got exception while listing files for {prefix=}", exc_info=True)
        raise


def _download_file_to_table(
    client: S3Client,
    bucket: str,
    filename: str,
    cols_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
) -> pa.Table:
    _logger.debug(f"Downloading parquet file {filename} into arrow table....")
    buffer = io.BytesIO()
    with _boto_lock_ctx():
        client.download_fileobj(Bucket=bucket, Key=filename, Fileobj=buffer)
        _logger.debug(f"Fetched parquet file {filename}. Reading into arrow...")
        buffer.seek(0)
        if env_var_bool("CHALK_REDSHIFT_POLARS_PARQUET"):
            tbl = read_parquet(buffer, use_pyarrow=False).to_arrow()
        else:
            tbl = pq.read_table(buffer)
        _logger.debug(f"Read parquet file {filename} into arrow: {tbl.nbytes=}, {tbl.num_rows=}. Converting columns...")
    # TODO (CHA-2232) Delete the `filename` from the bucket since we'll never look at it again

    # Infer results schema from table columns & map to correct dtype
    features = cols_to_features(tbl.column_names)
    results_schema = {v.root_fqn: v.converter.pyarrow_dtype for v in features.values()}
    tbl = tbl.select(list(features.keys())).rename_columns([x.root_fqn for x in features.values()])
    table = json_parse_and_cast(tbl, results_schema)
    _logger.debug(f"Finished downloading parquet file {filename}.")
    return table



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/query_values.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.common.v1 import query_log_pb2 as chalk_dot_common_dot_v1_dot_query__log__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n"chalk/common/v1/query_values.proto\x12\x0f\x63halk.common.v1\x1a\x1f\x63halk/common/v1/query_log.proto\x1a\x1fgoogle/protobuf/timestamp.proto"C\n\x1aOperationIdTableIdentifier\x12!\n\x0coperation_id\x18\x01 \x01(\tR\x0boperationId:\x02\x18\x01"y\n\x18TableNameTableIdentifier\x12\x1d\n\ntable_name\x18\x01 \x01(\tR\ttableName\x12:\n\x07\x66ilters\x18\x02 \x01(\x0b\x32 .chalk.common.v1.QueryLogFiltersR\x07\x66ilters:\x02\x18\x01"\xb1\x01\n\x17GetQueryValuesPageToken\x12J\n\x13query_timestamp_hwm\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x11queryTimestampHwm\x12(\n\x10operation_id_hwm\x18\x02 \x01(\tR\x0eoperationIdHwm\x12\x1c\n\nrow_id_hwm\x18\x03 \x01(\x03R\x08rowIdHwm:\x02\x18\x01"\xda\x04\n\x15GetQueryValuesRequest\x12\x65\n\x17operation_id_identifier\x18\x01 \x01(\x0b\x32+.chalk.common.v1.OperationIdTableIdentifierH\x00R\x15operationIdIdentifier\x12_\n\x15table_name_identifier\x18\x02 \x01(\x0b\x32).chalk.common.v1.TableNameTableIdentifierH\x00R\x13tableNameIdentifier\x12l\n%query_timestamp_lower_bound_inclusive\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR!queryTimestampLowerBoundInclusive\x12q\n%query_timestamp_upper_bound_exclusive\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x01R!queryTimestampUpperBoundExclusive\x88\x01\x01\x12\x1a\n\x08\x66\x65\x61tures\x18\x05 \x03(\tR\x08\x66\x65\x61tures\x12\x1b\n\tpage_size\x18\x07 \x01(\x05R\x08pageSize\x12\x1d\n\npage_token\x18\x08 \x01(\tR\tpageToken:\x02\x18\x01\x42\x12\n\x10table_identifierB(\n&_query_timestamp_upper_bound_exclusive"k\n\x16GetQueryValuesResponse\x12&\n\x0fnext_page_token\x18\x01 \x01(\tR\rnextPageToken\x12\x1a\n\x07parquet\x18\x02 \x01(\x0cH\x00R\x07parquet:\x02\x18\x01\x42\t\n\x07payloadB\x85\x01\n\x13\x63om.chalk.common.v1B\x10QueryValuesProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.query_values_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\020QueryValuesProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_OPERATIONIDTABLEIDENTIFIER"]._options = None
    _globals["_OPERATIONIDTABLEIDENTIFIER"]._serialized_options = b"\030\001"
    _globals["_TABLENAMETABLEIDENTIFIER"]._options = None
    _globals["_TABLENAMETABLEIDENTIFIER"]._serialized_options = b"\030\001"
    _globals["_GETQUERYVALUESPAGETOKEN"]._options = None
    _globals["_GETQUERYVALUESPAGETOKEN"]._serialized_options = b"\030\001"
    _globals["_GETQUERYVALUESREQUEST"]._options = None
    _globals["_GETQUERYVALUESREQUEST"]._serialized_options = b"\030\001"
    _globals["_GETQUERYVALUESRESPONSE"]._options = None
    _globals["_GETQUERYVALUESRESPONSE"]._serialized_options = b"\030\001"
    _globals["_OPERATIONIDTABLEIDENTIFIER"]._serialized_start = 121
    _globals["_OPERATIONIDTABLEIDENTIFIER"]._serialized_end = 188
    _globals["_TABLENAMETABLEIDENTIFIER"]._serialized_start = 190
    _globals["_TABLENAMETABLEIDENTIFIER"]._serialized_end = 311
    _globals["_GETQUERYVALUESPAGETOKEN"]._serialized_start = 314
    _globals["_GETQUERYVALUESPAGETOKEN"]._serialized_end = 491
    _globals["_GETQUERYVALUESREQUEST"]._serialized_start = 494
    _globals["_GETQUERYVALUESREQUEST"]._serialized_end = 1096
    _globals["_GETQUERYVALUESRESPONSE"]._serialized_start = 1098
    _globals["_GETQUERYVALUESRESPONSE"]._serialized_end = 1205
# @@protoc_insertion_point(module_scope)



import typing
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Protocol, Type, TypeVar, overload

from typing_extensions import Annotated

from chalk.features.feature_field import Feature
from chalk.features.feature_set import Features, is_feature_set_class, is_features_cls
from chalk.features.feature_wrapper import unwrap_feature

T = TypeVar("T")

if TYPE_CHECKING:

    class Primary(Protocol[T]):
        """Marks a feature as the primary feature for a feature class.

        Features named `id` on feature classes without an explicit primary
        feature are declared primary keys by default, and don't need to be
        marked with `Primary`.

        If you have primary key feature with a name other than `id`,
        you can use this marker to indicate the primary key.

        Examples
        --------
        >>> from chalk.features import features
        >>> from chalk import Primary
        >>> @features
        ... class User:
        ...     uid: Primary[int]
        """

        @overload
        def __get__(self, instance: None, owner: Any) -> Type[T]:
            ...

        @overload
        def __get__(self, instance: object, owner: Any) -> T:
            ...

        def __set__(self, instance: Any, value: T) -> None:
            ...

        def __delete__(self, instance: Any) -> None:
            ...

else:

    class Primary:
        """Marks a feature as the primary feature for a feature class.

        Features named `id` on feature classes without an explicit primary
        feature are declared primary keys by default, and don't need to be
        marked with `Primary`.

        If you have primary key feature with a name other than `id`,
        you can use this marker to indicate the primary key.

        Examples
        --------
        >>> from chalk.features import features
        >>> from chalk import Primary
        >>> @features
        ... class User:
        ...     username: Primary[str]
        """

        def __class_getitem__(cls, item: typing.Union[Type, str, int]):
            """
            Parameters
            ----------
            item
                The type of the feature value.

            Returns
            -------
            Annotated[item, "__chalk_primary__"]
                The type, with a special annotation indicating that it is a
                primary key.

            Examples
            --------
            >>> from chalk.features import features
            >>> from chalk import Primary
            >>> @features
            ... class User:
            ...     username: Primary[str]
            """
            return Annotated[item, "__chalk_primary__"]


def is_primary(f: Any) -> bool:
    """Determine whether a feature is a primary key.

    Parameters
    ----------
    f
        A feature (i.e. `User.email`)

    Raises
    ------
    TypeError
        If `f` is not a feature.

    Returns
    -------
    bool
        `True` if `f` is primary and `False` otherwise.

    Examples
    --------
    >>> from chalk.features import features
    >>> from chalk import Primary
    >>> @features
    ... class User:
    ...     uid: Primary[int]
    ...     email: str
    >>> assert is_primary(User.uid)
    >>> assert not is_primary(User.email)
    """
    return unwrap_feature(f).primary


def get_primary_feature(feature_class: typing.Union[Type[Features], typing.Iterable[Feature]]) -> Feature:
    """Returns the primary key feature of a features class or a set of features.

    Parameters
    ----------
    feature_class
        Either a features class or an iterable containing Chalk feature fields.

    Examples
    --------
    >>> from chalk.features import features
    >>> from chalk import Primary
    >>> @features
    ... class User:
    ...     uid: Primary[int]
    ...     email: str
    ...     first_name: str
    ...     last_name: str
    >>> assert get_primary_feature(User).fqn == "user.uid"
    >>> assert get_primary_feature(User).primary
    >>> assert get_primary_feature(Features[User.uid, User.email]).fqn == "user.uid"
    >>> assert get_primary_feature(Features[User.uid, User.email]).primary

    :param feature_class: Either a features class or an iterable containing Chalk feature fields.
    :return: The primary feature for the given feature class/set. Raises an exception if there is no primary key present.
    """
    if is_feature_set_class(feature_class):
        if feature_class.__chalk_primary__ is not None:
            return feature_class.__chalk_primary__
        raise ValueError(f"Feature set {feature_class} does not contain a primary feature.")
    elif is_features_cls(feature_class):
        for f in feature_class.features:
            if is_primary(f):
                return f
        raise ValueError(f"Feature set {feature_class} does not contain a primary feature.")
    elif isinstance(feature_class, Iterable):
        for f in feature_class:
            if is_primary(unwrap_feature(f)):
                return f
        raise ValueError(f"Feature set {feature_class} does not contain a primary feature.")
    else:
        raise ValueError(f"Object {feature_class} is not a feature set.")



from __future__ import annotations

import asyncio
import collections.abc
import contextlib
import functools
import queue
import threading
from concurrent.futures import Future, ThreadPoolExecutor
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Coroutine,
    Generic,
    Iterable,
    Mapping,
    Sequence,
    TypeVar,
    cast,
    overload,
)

import aiotools
from typing_extensions import ParamSpec

from chalk.utils.tracing import safe_activate_trace_context, safe_current_trace_context

T = TypeVar("T")
P = ParamSpec("P")
K = TypeVar("K")


@contextlib.asynccontextmanager
async def async_null_context(obj: T) -> AsyncIterator[T]:
    yield obj


async def async_enumerate(iterable: AsyncIterator[T] | AsyncIterable[T]):
    i = 0
    async for x in iterable:
        yield i, x
        i += 1


_RUNNING_TASKS: set[asyncio.Task[Any]] = set()


def run_coroutine_fn_threadsafe(
    loop: asyncio.AbstractEventLoop,
    coro_fn: Callable[P, Coroutine[Any, Any, T]],
    /,
    *args: P.args,
    **kwargs: P.kwargs,
) -> Future[T]:
    """Similar to asyncio.run_coroutine_threadsafe(), but it constructs the coroutine inside the main event loop thread.
    This helps eliminate errors related to constructing coroutines that will never be awaited if the event loop is closed before
    the coroutine could be executed"""
    ans: Future[T] = Future()

    current_trace_context = safe_current_trace_context()

    def _run_on_event_loop_thread():
        if ans.set_running_or_notify_cancel():

            @functools.wraps(coro_fn)
            async def wrapped_with_context(*args: P.args, **kwargs: P.kwargs):
                try:
                    safe_activate_trace_context(current_trace_context)
                    res = await coro_fn(*args, **kwargs)
                except BaseException as exc:
                    ans.set_exception(exc)
                else:
                    ans.set_result(res)

            t = asyncio.create_task(wrapped_with_context(*args, **kwargs))
            _RUNNING_TASKS.add(t)
            t.add_done_callback(_RUNNING_TASKS.remove)
        else:
            ans.set_exception(RuntimeError("Future was cancelled"))

    loop.call_soon_threadsafe(_run_on_event_loop_thread)
    return ans


def _put_on_queue(
    loop: asyncio.AbstractEventLoop,
    q: asyncio.Queue[T | ellipsis],
    finished_event: threading.Event,
    x: T | ellipsis,
):
    if finished_event.is_set():
        return
    run_coroutine_fn_threadsafe(loop, q.put, x).result()


def _yield_to_queue(
    loop: asyncio.AbstractEventLoop,
    q: asyncio.Queue[T | ellipsis],
    finished_event: threading.Event,
    it: Iterable[T],
):
    try:
        for x in it:
            _put_on_queue(loop, q, finished_event, x)
    finally:
        _put_on_queue(loop, q, finished_event, ...)


class to_async_iterable(Generic[T]):
    """Runs a blocking iterator in an executor, and yields batches asynchronously as they become available.

    This function-like class runs the generator in a separate thread, and uses a queue to share results between the background thread
    and the event loop. This approach ensures that the generator context is spun with the correct parent
    """

    @overload
    def __init__(
        self,
        iterable: Iterable[T],
        executor: ThreadPoolExecutor | None = None,
        loop: None = None,
    ):
        ...

    @overload
    def __init__(
        self,
        iterable: Iterable[T],
        executor: ThreadPoolExecutor,
        loop: asyncio.AbstractEventLoop,
    ):
        ...

    def __init__(
        self,
        iterable: Iterable[T],
        executor: ThreadPoolExecutor | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ):
        super().__init__()
        self._q: asyncio.Queue[T | ellipsis] = asyncio.Queue(maxsize=2)
        self._finished_event = threading.Event()
        if isinstance(iterable, (list, tuple, set, frozenset)):
            # For simple containers we can iterate directly; no need to use a queue / task
            self._task = None
            self._loop = None
            self._iterator = iter(iterable)
        else:
            self._iterator = None
            if loop is None:
                self._loop = asyncio.get_running_loop()
                self._task = self._loop.run_in_executor(
                    executor,
                    _yield_to_queue,
                    self._loop,
                    self._q,
                    self._finished_event,
                    iterable,
                )
            else:
                assert executor is not None, "If the loop is provided, an executor must also be provided"
                self._loop = loop
                self._task = executor.submit(
                    _yield_to_queue,
                    self._loop,
                    self._q,
                    self._finished_event,
                    iterable,
                )

    def __aiter__(self):
        return self

    async def __anext__(self) -> T:
        if self._iterator is not None:
            element = next(self._iterator, ...)
            if element is ...:
                raise StopAsyncIteration
            else:
                return element
        assert self._task is not None
        try:
            item = await self._q.get()
        except asyncio.CancelledError:
            # If the task is cancelled, we still need to add something to the queue so the .get() call above doesn't hang
            # When the worker task finishes, it will either a) be before any code below runs, which is fine
            # because it is idempotent, or b) be after the the finished event is set, in which case we won't attempt to
            # append anything
            self._finished_event.set()
            try:
                self._q.put_nowait(...)
            except queue.Full:
                pass
            raise
        if item is ...:
            await asyncio.wrap_future(self._task, loop=self._loop)
            raise StopAsyncIteration
        return item

    def __del__(self):
        # If there was an exception, then the queue might still have an element in it
        # We'll first set the finished_event, and then we'll drain the queue. This will allow the
        # _yield_to_queue to successfully put another element in it, if _safe_put is blocked on q.put(...)
        # Because the finished_event is already set, it will not attempt to add yet another element, which would
        # block forever, because we aren't draining the queue any more
        # If there was no exception, then the queue should already be empty, and the task complete
        self._finished_event.set()
        try:
            if self._loop is not None:
                self._loop.call_soon_threadsafe(self._wrapped_get, self._q)
        except RuntimeError:
            pass

    @staticmethod
    def _wrapped_get(q: asyncio.Queue[T | ellipsis]):
        # Staticmethod so this doesn't hold a reference to self
        try:
            q.get_nowait()
        except asyncio.QueueEmpty:
            pass


@overload
async def wait_for_futures(
    futures: Iterable[Coroutine[Any, Any, T]],
) -> Sequence[T]:
    ...


@overload
async def wait_for_futures(
    futures: Mapping[K, Coroutine[Any, Any, T]],
) -> Mapping[K, T]:
    ...


async def wait_for_futures(
    futures: Mapping[K, Coroutine[Any, Any, T]] | Iterable[Coroutine[Any, Any, T]],
) -> Mapping[K, T] | Sequence[T]:
    """Run all the provided coroutines, and at the end, raise exceptions that might have occurred.

    This method will greedily cancel any coroutines if there is an exception."""
    if isinstance(futures, collections.abc.Mapping):
        was_mapping = True
        futs_map = cast(Mapping[K, Coroutine[Any, Any, T]], futures)
    else:
        was_mapping = False
        futs_map = {cast(K, i): fut for (i, fut) in enumerate(futures)}

    if len(futs_map) == 0:
        if was_mapping:
            return {}
        else:
            return []
    ans: dict[K, T] = {}
    if len(futs_map) == 1:
        # If a single future, skipping the asyncio.gather because spawning asyncio tasks is slow
        for k, v in futs_map.items():
            ans[k] = await v
    else:
        async with aiotools.TaskGroup() as tg:
            tasks = tuple((k, tg.create_task(v)) for (k, v) in futs_map.items())
        for task_name, t in tasks:
            ans[task_name] = t.result()
    if not was_mapping:
        return [ans[cast(K, i)] for i in range(len(ans))]
    return ans



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class CommandLineInterfaceVersion(_message.Message):
    __slots__ = ("version", "download_url", "os", "arch", "generation")
    VERSION_FIELD_NUMBER: _ClassVar[int]
    DOWNLOAD_URL_FIELD_NUMBER: _ClassVar[int]
    OS_FIELD_NUMBER: _ClassVar[int]
    ARCH_FIELD_NUMBER: _ClassVar[int]
    GENERATION_FIELD_NUMBER: _ClassVar[int]
    version: str
    download_url: str
    os: str
    arch: str
    generation: int
    def __init__(
        self,
        version: _Optional[str] = ...,
        download_url: _Optional[str] = ...,
        os: _Optional[str] = ...,
        arch: _Optional[str] = ...,
        generation: _Optional[int] = ...,
    ) -> None: ...

class GetVersionsRequest(_message.Message):
    __slots__ = ("os", "arch")
    OS_FIELD_NUMBER: _ClassVar[int]
    ARCH_FIELD_NUMBER: _ClassVar[int]
    os: str
    arch: str
    def __init__(self, os: _Optional[str] = ..., arch: _Optional[str] = ...) -> None: ...

class GetVersionsResponse(_message.Message):
    __slots__ = ("versions", "latest", "nightly", "minimum")
    VERSIONS_FIELD_NUMBER: _ClassVar[int]
    LATEST_FIELD_NUMBER: _ClassVar[int]
    NIGHTLY_FIELD_NUMBER: _ClassVar[int]
    MINIMUM_FIELD_NUMBER: _ClassVar[int]
    versions: _containers.RepeatedCompositeFieldContainer[CommandLineInterfaceVersion]
    latest: CommandLineInterfaceVersion
    nightly: CommandLineInterfaceVersion
    minimum: str
    def __init__(
        self,
        versions: _Optional[_Iterable[_Union[CommandLineInterfaceVersion, _Mapping]]] = ...,
        latest: _Optional[_Union[CommandLineInterfaceVersion, _Mapping]] = ...,
        nightly: _Optional[_Union[CommandLineInterfaceVersion, _Mapping]] = ...,
        minimum: _Optional[str] = ...,
    ) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/common/v1/feature_values.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.chart.v1 import densetimeserieschart_pb2 as chalk_dot_chart_dot_v1_dot_densetimeserieschart__pb2
from chalk._gen.chalk.common.v1 import chart_pb2 as chalk_dot_common_dot_v1_dot_chart__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n$chalk/common/v1/feature_values.proto\x12\x0f\x63halk.common.v1\x1a)chalk/chart/v1/densetimeserieschart.proto\x1a\x1b\x63halk/common/v1/chart.proto\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xff\x01\n\x1cGetFeatureValuesChartRequest\x12\x10\n\x03\x66qn\x18\x01 \x01(\tR\x03\x66qn\x12K\n\x0c\x61ggregate_by\x18\x02 \x03(\x0e\x32(.chalk.common.v1.FeatureValueAggregationR\x0b\x61ggregateBy\x12(\n\rwindow_period\x18\x03 \x01(\tH\x00R\x0cwindowPeriod\x88\x01\x01\x12\x19\n\x08start_ms\x18\x04 \x01(\x03R\x07startMs\x12\x1a\n\x06\x65nd_ms\x18\x05 \x01(\x03H\x01R\x05\x65ndMs\x88\x01\x01:\x02\x18\x01\x42\x10\n\x0e_window_periodB\t\n\x07_end_ms"Q\n\x1dGetFeatureValuesChartResponse\x12,\n\x05\x63hart\x18\x01 \x01(\x0b\x32\x16.chalk.common.v1.ChartR\x05\x63hart:\x02\x18\x01"\xaf\x03\n&GetFeatureValuesTimeSeriesChartRequest\x12\x10\n\x03\x66qn\x18\x01 \x01(\tR\x03\x66qn\x12K\n\x0c\x61ggregate_by\x18\x02 \x03(\x0e\x32(.chalk.common.v1.FeatureValueAggregationR\x0b\x61ggregateBy\x12\x43\n\rwindow_period\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationH\x00R\x0cwindowPeriod\x88\x01\x01\x12V\n\x19start_timestamp_inclusive\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x17startTimestampInclusive\x12W\n\x17\x65nd_timestamp_exclusive\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x01R\x15\x65ndTimestampExclusive\x88\x01\x01:\x02\x18\x01\x42\x10\n\x0e_window_periodB\x1a\n\x18_end_timestamp_exclusive"3\n\'GetFeatureValuesTimeSeriesChartResponse:\x02\x18\x01J\x04\x08\x01\x10\x02"\xb1\x03\n(GetFeatureValuesTimeSeriesChartV2Request\x12\x10\n\x03\x66qn\x18\x01 \x01(\tR\x03\x66qn\x12K\n\x0c\x61ggregate_by\x18\x02 \x03(\x0e\x32(.chalk.common.v1.FeatureValueAggregationR\x0b\x61ggregateBy\x12\x43\n\rwindow_period\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationH\x00R\x0cwindowPeriod\x88\x01\x01\x12V\n\x19start_timestamp_inclusive\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x17startTimestampInclusive\x12W\n\x17\x65nd_timestamp_exclusive\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x01R\x15\x65ndTimestampExclusive\x88\x01\x01:\x02\x18\x01\x42\x10\n\x0e_window_periodB\x1a\n\x18_end_timestamp_exclusive"k\n)GetFeatureValuesTimeSeriesChartV2Response\x12:\n\x05\x63hart\x18\x01 \x01(\x0b\x32$.chalk.chart.v1.DenseTimeSeriesChartR\x05\x63hart:\x02\x18\x01*\xaa\x04\n\x17\x46\x65\x61tureValueAggregation\x12)\n%FEATURE_VALUE_AGGREGATION_UNSPECIFIED\x10\x00\x12+\n\'FEATURE_VALUE_AGGREGATION_UNIQUE_VALUES\x10\x01\x12\x30\n,FEATURE_VALUE_AGGREGATION_TOTAL_OBSERVATIONS\x10\x02\x12-\n)FEATURE_VALUE_AGGREGATION_NULL_PERCENTAGE\x10\x03\x12\'\n#FEATURE_VALUE_AGGREGATION_MAX_VALUE\x10\x04\x12\'\n#FEATURE_VALUE_AGGREGATION_MIN_VALUE\x10\x05\x12%\n!FEATURE_VALUE_AGGREGATION_AVERAGE\x10\x06\x12*\n&FEATURE_VALUE_AGGREGATION_UNIQUE_PKEYS\x10\x07\x12!\n\x1d\x46\x45\x41TURE_VALUE_AGGREGATION_P95\x10\x08\x12!\n\x1d\x46\x45\x41TURE_VALUE_AGGREGATION_P75\x10\t\x12!\n\x1d\x46\x45\x41TURE_VALUE_AGGREGATION_P50\x10\n\x12!\n\x1d\x46\x45\x41TURE_VALUE_AGGREGATION_P25\x10\x0b\x12!\n\x1d\x46\x45\x41TURE_VALUE_AGGREGATION_P05\x10\x0c\x1a\x02\x18\x01\x42\x87\x01\n\x13\x63om.chalk.common.v1B\x12\x46\x65\x61tureValuesProtoP\x01\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43halk.Common.V1\xca\x02\x0f\x43halk\\Common\\V1\xe2\x02\x1b\x43halk\\Common\\V1\\GPBMetadata\xea\x02\x11\x43halk::Common::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.common.v1.feature_values_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.common.v1B\022FeatureValuesProtoP\001\242\002\003CCX\252\002\017Chalk.Common.V1\312\002\017Chalk\\Common\\V1\342\002\033Chalk\\Common\\V1\\GPBMetadata\352\002\021Chalk::Common::V1"
    _globals["_FEATUREVALUEAGGREGATION"]._options = None
    _globals["_FEATUREVALUEAGGREGATION"]._serialized_options = b"\030\001"
    _globals["_GETFEATUREVALUESCHARTREQUEST"]._options = None
    _globals["_GETFEATUREVALUESCHARTREQUEST"]._serialized_options = b"\030\001"
    _globals["_GETFEATUREVALUESCHARTRESPONSE"]._options = None
    _globals["_GETFEATUREVALUESCHARTRESPONSE"]._serialized_options = b"\030\001"
    _globals["_GETFEATUREVALUESTIMESERIESCHARTREQUEST"]._options = None
    _globals["_GETFEATUREVALUESTIMESERIESCHARTREQUEST"]._serialized_options = b"\030\001"
    _globals["_GETFEATUREVALUESTIMESERIESCHARTRESPONSE"]._options = None
    _globals["_GETFEATUREVALUESTIMESERIESCHARTRESPONSE"]._serialized_options = b"\030\001"
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2REQUEST"]._options = None
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2REQUEST"]._serialized_options = b"\030\001"
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2RESPONSE"]._options = None
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2RESPONSE"]._serialized_options = b"\030\001"
    _globals["_FEATUREVALUEAGGREGATION"]._serialized_start = 1568
    _globals["_FEATUREVALUEAGGREGATION"]._serialized_end = 2122
    _globals["_GETFEATUREVALUESCHARTREQUEST"]._serialized_start = 195
    _globals["_GETFEATUREVALUESCHARTREQUEST"]._serialized_end = 450
    _globals["_GETFEATUREVALUESCHARTRESPONSE"]._serialized_start = 452
    _globals["_GETFEATUREVALUESCHARTRESPONSE"]._serialized_end = 533
    _globals["_GETFEATUREVALUESTIMESERIESCHARTREQUEST"]._serialized_start = 536
    _globals["_GETFEATUREVALUESTIMESERIESCHARTREQUEST"]._serialized_end = 967
    _globals["_GETFEATUREVALUESTIMESERIESCHARTRESPONSE"]._serialized_start = 969
    _globals["_GETFEATUREVALUESTIMESERIESCHARTRESPONSE"]._serialized_end = 1020
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2REQUEST"]._serialized_start = 1023
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2REQUEST"]._serialized_end = 1456
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2RESPONSE"]._serialized_start = 1458
    _globals["_GETFEATUREVALUESTIMESERIESCHARTV2RESPONSE"]._serialized_end = 1565
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/metrics.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1d\x63halk/server/v1/metrics.proto\x12\x0f\x63halk.server.v1\x1a\x1f\x63halk/auth/v1/permissions.proto"\x88\x01\n GetOverviewSummaryMetricsRequest\x12$\n\x0brange_start\x18\x01 \x01(\tH\x00R\nrangeStart\x88\x01\x01\x12 \n\trange_end\x18\x02 \x01(\tH\x01R\x08rangeEnd\x88\x01\x01\x42\x0e\n\x0c_range_startB\x0c\n\n_range_end"A\n\x15OverviewSummaryMetric\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x14\n\x05value\x18\x02 \x01(\x02R\x05value"e\n!GetOverviewSummaryMetricsResponse\x12@\n\x07metrics\x18\x01 \x03(\x0b\x32&.chalk.server.v1.OverviewSummaryMetricR\x07metrics2\x9a\x01\n\x0eMetricsService\x12\x87\x01\n\x19GetOverviewSummaryMetrics\x12\x31.chalk.server.v1.GetOverviewSummaryMetricsRequest\x1a\x32.chalk.server.v1.GetOverviewSummaryMetricsResponse"\x03\x80}\x06\x42\x95\x01\n\x13\x63om.chalk.server.v1B\x0cMetricsProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.metrics_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\014MetricsProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_METRICSSERVICE"].methods_by_name["GetOverviewSummaryMetrics"]._options = None
    _globals["_METRICSSERVICE"].methods_by_name["GetOverviewSummaryMetrics"]._serialized_options = b"\200}\006"
    _globals["_GETOVERVIEWSUMMARYMETRICSREQUEST"]._serialized_start = 84
    _globals["_GETOVERVIEWSUMMARYMETRICSREQUEST"]._serialized_end = 220
    _globals["_OVERVIEWSUMMARYMETRIC"]._serialized_start = 222
    _globals["_OVERVIEWSUMMARYMETRIC"]._serialized_end = 287
    _globals["_GETOVERVIEWSUMMARYMETRICSRESPONSE"]._serialized_start = 289
    _globals["_GETOVERVIEWSUMMARYMETRICSRESPONSE"]._serialized_end = 390
    _globals["_METRICSSERVICE"]._serialized_start = 393
    _globals["_METRICSSERVICE"]._serialized_end = 547
# @@protoc_insertion_point(module_scope)



from typing import TYPE_CHECKING, Any, Protocol, Set, Type, TypeVar, overload

from chalk.features import Feature
from chalk.features.feature_wrapper import unwrap_feature

T = TypeVar("T")

if TYPE_CHECKING:

    class Last(Protocol[T]):
        @overload
        def __get__(self, instance: None, owner: Any) -> Type[T]:
            ...

        @overload
        def __get__(self, instance: object, owner: Any) -> T:
            ...

        def __set__(self, instance: Any, value: T) -> None:
            ...

        def __delete__(self, instance: Any) -> None:
            ...

else:

    class Last:
        """Marks a feature as the last observed value of a feature.

        Examples
        --------
        >>> from chalk.features import features, online
        >>> @features
        ... class User:
        ...     id: int
        ...     model_score: float
        >>> @online
        ... def get_score(
        ...     s: Last[User.model_score]
        ... ) -> User.model_score:
        ...     return s * 1.2
        """

        registry: Set[Feature]

        def __class_getitem__(cls, item):
            return unwrap_feature(item).as_last()



from chalk._gen.chalk.common.v1 import online_query_pb2 as _online_query_pb2
from chalk._gen.chalk.engine.v1 import plan_pb2 as _plan_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetPlanRequest(_message.Message):
    __slots__ = ("online_query_request",)
    ONLINE_QUERY_REQUEST_FIELD_NUMBER: _ClassVar[int]
    online_query_request: _online_query_pb2.OnlineQueryRequest
    def __init__(
        self, online_query_request: _Optional[_Union[_online_query_pb2.OnlineQueryRequest, _Mapping]] = ...
    ) -> None: ...

class GetPlanResponse(_message.Message):
    __slots__ = ("plan",)
    PLAN_FIELD_NUMBER: _ClassVar[int]
    plan: _plan_pb2.Plan
    def __init__(self, plan: _Optional[_Union[_plan_pb2.Plan, _Mapping]] = ...) -> None: ...

class ExecuteQueryRequest(_message.Message):
    __slots__ = ("online_query_request",)
    ONLINE_QUERY_REQUEST_FIELD_NUMBER: _ClassVar[int]
    online_query_request: _online_query_pb2.OnlineQueryRequest
    def __init__(
        self, online_query_request: _Optional[_Union[_online_query_pb2.OnlineQueryRequest, _Mapping]] = ...
    ) -> None: ...

class ExecuteQueryResponse(_message.Message):
    __slots__ = ("online_query_response",)
    ONLINE_QUERY_RESPONSE_FIELD_NUMBER: _ClassVar[int]
    online_query_response: _online_query_pb2.OnlineQueryResponse
    def __init__(
        self, online_query_response: _Optional[_Union[_online_query_pb2.OnlineQueryResponse, _Mapping]] = ...
    ) -> None: ...



from __future__ import annotations

import datetime
import functools
import re
from typing import Any

import google.cloud.bigquery.dbapi._helpers
import google.cloud.bigquery.dbapi.cursor
import google.cloud.bigquery.exceptions
import google.cloud.bigquery.query
import sqlalchemy.sql.functions
import sqlalchemy.sql.visitors
from sqlalchemy import BINARY, Column, DefaultClause, LargeBinary, Table, TypeDecorator
from sqlalchemy.engine import Dialect
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import CreateIndex, CreateTable, DefaultClause
from sqlalchemy.sql import FromClause, Selectable, Update
from sqlalchemy.sql.compiler import DDLCompiler
from sqlalchemy.sql.ddl import DropTable
from sqlalchemy.sql.elements import BinaryExpression, BindParameter, Cast, ColumnElement
from sqlalchemy.sql.operators import json_path_getitem_op
from sqlalchemy.sql.selectable import CTE, CompoundSelectState, SelectState
from sqlalchemy.sql.sqltypes import STRINGTYPE, NullType, _Binary  # pyright: ignore[reportPrivateUsage]
from sqlalchemy.types import JSON, DateTime, Interval
from sqlalchemy_bigquery import STRUCT, BigQueryDialect
from sqlalchemy_bigquery.base import BigQueryCompiler, BigQueryDDLCompiler, BigQueryTypeCompiler, process_string_literal


def _compile_cast(element: Cast, compiler: BigQueryCompiler, **kw: Any):
    if isinstance(element.type, JSON):
        assert isinstance(element.clause, ColumnElement)
        processed_text = compiler.process(element.clause, **kw)
        return f"PARSE_JSON({processed_text})"
    return compiler.visit_cast(element, **kw)


def _compile_drop_table(element: DropTable, compiler: BigQueryDDLCompiler, **kw: Any):
    element.if_exists = True
    return compiler.visit_drop_table(element, **kw)


def _format_datetime_as_iso(val: datetime.datetime):
    """Format a datetime as standardized ISO string, independent of platform"""
    # See https://bugs.python.org/issue13305 -- `datetime.datetime.min.isoformat()` is incorrect on linux
    # It returns 1-01-01T00:00:00 instead of 0001-01-01T00:00:00
    time_format = val.time().isoformat()
    return f"{val.year:04d}-{val.month:02d}-{val.day:02d}T{time_format}"


def _compiles_bind_param(element: BindParameter, compiler: BigQueryCompiler, **kw: Any):
    dialect = compiler.dialect
    assert isinstance(dialect, BigQueryDialect)
    if isinstance(element.type, JSON.JSONPathType):
        if element.value != []:
            raise NotImplementedError("Need to implement full JSON path parsing")
        return '"$"'  # JSON path for everything
    assert isinstance(dialect.type_compiler, BigQueryTypeCompiler)
    # FIXME: This fixes chalk sql (CHA-1356)
    if isinstance(element.type, NullType):
        return compiler.visit_bindparam(element, **kw)
    element_type = dialect.type_compiler.process(element.type)

    # FIXME: Hacking to fix a demo. Issue is that our chalk sql resolver bind params come through with this type.
    if element_type == "INTERVAL" and isinstance(element.value, datetime.timedelta):
        dt = element.value

        seconds = int(dt.seconds)
        if dt.microseconds > 0:
            seconds += float(dt.microseconds) / 1e6
        iso8601 = f'CAST("P{int(abs(dt.days))}DT{seconds}S" AS INTERVAL)'
        if dt.seconds < 0:
            iso8601 = f"- {iso8601}"
        return iso8601
    if element_type == "DATETIME" or element_type == "TIMESTAMP":
        if isinstance(element.type, TypeDecorator):
            element.value = element.type.process_bind_param(element.value, compiler.dialect)
        if isinstance(element.value, str):
            if element.value.lower() in ("infinity", "inf", "+inf", "+infinity"):
                element.value = datetime.datetime.max
                if element_type == "TIMESTAMP":
                    element.value = element.value.replace(tzinfo=datetime.timezone.utc)
            elif element.value.lower() in ("-infinity", "-inf"):
                element.value = datetime.datetime.min
                if element_type == "TIMESTAMP":
                    element.value = element.value.replace(tzinfo=datetime.timezone.utc)
        if isinstance(element.value, datetime.datetime):
            if element_type == "TIMESTAMP" and element.value.tzinfo is None:
                raise ValueError(f"TIMESTAMP types must have a timezone, but '{element.value}' is unzoned")
            if element_type == "DATETIME" and element.value.tzinfo is not None:
                raise ValueError(f"DATETIME types do not have a timezone, but '{element.value}' is zoned")
            element.value = _format_datetime_as_iso(element.value)
    return compiler.visit_bindparam(element, **kw)


def _compile_interval(element: Interval, compiler: BigQueryTypeCompiler, **kw: Any):
    return "INTERVAL"


def _compile_struct(element: STRUCT, compiler: BigQueryTypeCompiler, **kw: Any):
    fields = ", ".join(
        f"`{name}` {compiler.process(type_, **kw)}"
        for name, type_ in element._STRUCT_fields  # pyright: ignore[reportPrivateUsage]
    )
    return f"STRUCT<{fields}>"


def _compile_create_index(element: CreateIndex, compiler: BigQueryDDLCompiler, **kw: Any):
    if element.element.dialect_options.get(compiler.dialect.name, {}).get("skip_index", False):
        # No-op query to prevent creation of the index
        return "SELECT 1"
    # Otherwise, best-effort to create the index
    sql_str = compiler.visit_create_index(element, **kw)
    return sql_str.replace("CREATE INDEX", "CREATE SEARCH INDEX")


def _compile_create_table(element: CreateTable, compiler: BigQueryDDLCompiler, **kw: Any):
    element.if_not_exists = True
    options = []
    dialect_options = element.element.dialect_options.get(compiler.dialect.name, {})
    if (bigquery_expiration := dialect_options.get("expiration")) is not None:
        assert isinstance(bigquery_expiration, datetime.datetime), "bigquery_expiration must be a datetime instance"
        options.append(f'expiration_timestamp=TIMESTAMP "{_format_datetime_as_iso(bigquery_expiration)}"')
    if element.element.comment is not None:
        options.append(f"description={process_string_literal(element.element.comment)}")
    if (bigquery_labels := dialect_options.get("labels")) is not None:
        assert isinstance(bigquery_labels, dict), "bigquery_labels must be a dict"
        labels = (
            "["
            + ",".join(
                "(" + process_string_literal(str(k)) + "," + process_string_literal(str(v)) + ")"
                for (k, v) in bigquery_labels.items()
            )
            + "]"
        )
        options.append(f"labels={labels}")
    partition = ""
    if (bigquery_partition := dialect_options.get("partition")) is not None:
        assert isinstance(bigquery_partition, str), "bigquery_partition must be string"
        partition = f"PARTITION BY {bigquery_partition}"
    cluster = ""
    if (bigquery_cluster := dialect_options.get("cluster")) is not None:
        assert isinstance(bigquery_cluster, (list, tuple)), "bigquery_cluster must a list of tuple of cluster columns"
        cluster = f"CLUSTER BY " + ", ".join(f"`{x}`" for x in bigquery_cluster)
    # Temporarily monkeypatching the compiler to skip the `post_create_table`, since we manually specify the options below
    original_post_create_table = compiler.post_create_table
    compiler.post_create_table = lambda *args, **kwargs: ""
    try:
        sql_str = compiler.visit_create_table(element, **kw)
    finally:
        compiler.post_create_table = original_post_create_table

    return "{sql_str} {partition} {cluster} {options}".format(
        sql_str=sql_str,
        partition=partition,
        cluster=cluster,
        options="" if len(options) == 0 else f"OPTIONS({', '.join(options)})",
    )


def _patched_get_column_default_string(self: DDLCompiler, column: Column):
    # Patching this function to include a cast
    if isinstance(column.server_default, DefaultClause):
        if isinstance(column.server_default.arg, str):
            return self.sql_compiler.render_literal_value(column.server_default.arg, STRINGTYPE)
        else:
            expected_type = self.type_compiler.process(column.type)
            return (
                f"CAST({self.sql_compiler.process(column.server_default.arg, literal_binds=True)} as {expected_type})"
            )
    else:
        return None


def _compile_func_now(element: sqlalchemy.sql.functions.now, compiler: BigQueryCompiler, **kw: Any):
    return "CURRENT_TIMESTAMP()"


def _compiles_binary_expression(element: BinaryExpression, compiler: BigQueryCompiler, **kw: Any):
    if element.operator == json_path_getitem_op:
        lhs = compiler.process(element.left, **kw)
        rhs = compiler.process(element.right, **kw)
        dialect = compiler.dialect
        element_type = dialect.type_compiler.process(element.type)
        assert isinstance(dialect, BigQueryDialect)
        return f"CAST(JSON_VALUE({lhs}, {rhs}) AS {element_type})"
    return compiler.visit_binary(element, **kw)


def _compile_datetime(element: DateTime, compiler: BigQueryTypeCompiler, **kw: Any):
    if element.timezone:
        return "TIMESTAMP"
    else:
        return "DATETIME"


def _patched_known_tables(self: BigQueryCompiler):
    # Copied from https://github.com/googleapis/python-bigquery-sqlalchemy/blob/94d113dcb0b763c74ef56842d1fbb0b046edfd11/sqlalchemy_bigquery/base.py#L257
    # and patched to work with compound select statements
    known_tables = set()
    compile_states = []
    if isinstance(self.compile_state, CompoundSelectState):
        for select in self.compile_state.statement.selects:
            compile_state = SelectState(select, self)
            compile_states.append(compile_state)
    else:
        compile_states.append(self.compile_state)

    for compile_state in compile_states:
        for from_ in compile_state.froms:
            if isinstance(from_, Table):
                known_tables.add(from_.name)
            elif isinstance(from_, CTE):
                for column in from_.original.selected_columns:  # type: ignore
                    table = getattr(column, "table", None)
                    if table is not None:
                        known_tables.add(table.name)

    return known_tables


def _patched_update_from(
    self: BigQueryCompiler,
    update_stmt: Update,
    from_table: FromClause,
    extra_froms: list[Selectable],
    from_hints: Any,
    **kw: Any,
):
    # Copied from https://github.com/sqlalchemy/sqlalchemy/blob/e15d00303580a78d6e54da896406ba1ec3685ca1/lib/sqlalchemy/dialects/postgresql/base.py#L1991
    kw["asfrom"] = True
    return "FROM " + ", ".join(
        t._compiler_dispatch(self, fromhints=from_hints, **kw)  # pyright: ignore -- this _compiler_dispatch exists
        for t in extra_froms
    )


def _patch_bigquery_dbapi():
    """Monkeypatching the `extract_types` function to handle struct field names with backticks"""
    # See https://github.com/googleapis/python-bigquery/blob/0bf95460866089c8e955c97ae02f2fa443e1ef62/google/cloud/bigquery/dbapi/cursor.py#L495
    # for the original source
    extract_types_sub = re.compile(
        r"""
            (%*)          # Extra %s.  We'll deal with these in the replacement code

            %             # Beginning of replacement, %s, %(...)s

            (?:\(         # Begin of optional name and/or type
            ([^:)]*)      # name
            (?::          # ':' introduces type
            (             # start of type group
                [a-zA-Z0-9_<>`, ]+ # First part, no parens  # <------------------ THIS IS THE LINE THAT DIFFERS

                (?:               # start sets of parens + non-paren text
                \([0-9 ,]+\)      # comma-separated groups of digits in parens
                                    # (e.g. string(10))
                (?=[, >)])        # Must be followed by ,>) or space
                [a-zA-Z0-9<>, ]*  # Optional non-paren chars
                )*                # Can be zero or more of parens and following text
            )             # end of type group
            )?            # close type clause ":type"
            \))?          # End of optional name and/or type

            s             # End of replacement
            """,
        re.VERBOSE,
    ).sub

    google.cloud.bigquery.dbapi.cursor._extract_types.__defaults__ = (  # pyright: ignore[reportPrivateUsage]
        extract_types_sub,
    )

    complex_query_parameter_parse = re.compile(
        r"""
        \s*
        (ARRAY|STRUCT|RECORD)  # Type
        \s*
        <([A-Z0-9_<>` ,()]+)>   # Subtype(s)    # <------------------ THIS IS THE LINE THAT DIFFERS
        \s*$
        """,
        re.IGNORECASE | re.VERBOSE,
    ).match
    google.cloud.bigquery.dbapi._helpers._parse_type.__defaults__ = (  # pyright: ignore[reportPrivateUsage]
        complex_query_parameter_parse,
    )

    parse_struct_field = re.compile(
        r"""
        (?:`?(\w+)`?\s+)    # field name   # <------------------ THIS IS THE LINE THAT DIFFERS
        ([A-Z0-9<> ,()]+)  # Field type
        $""",
        re.VERBOSE | re.IGNORECASE,
    ).match
    google.cloud.bigquery.dbapi._helpers._parse_struct_fields.__defaults__ = (  # pyright: ignore[reportPrivateUsage]
        parse_struct_field,
    )


class BQBINARY(BINARY):
    def literal_processor(self, dialect: Dialect):
        return lambda value: repr(value.replace(b"%", b"%%"))


class BQLargeBinary(LargeBinary):
    def literal_processor(self, dialect: Dialect):
        return lambda value: repr(value.replace(b"%", b"%%"))


@functools.lru_cache(None)
def register_bigquery_compiler_hooks():
    colspecs = dict(BigQueryDialect.colspecs)
    colspecs.pop(_Binary)
    colspecs[BINARY] = BQBINARY
    colspecs[LargeBinary] = BQLargeBinary
    BigQueryDialect.colspecs = colspecs
    BigQueryCompiler.update_from_clause = _patched_update_from
    BigQueryCompiler._known_tables = _patched_known_tables  # pyright: ignore[reportPrivateUsage]
    BigQueryDDLCompiler.get_column_default_string = _patched_get_column_default_string  # type: ignore -- stubs are incorrect
    compiles(BindParameter, "bigquery")(_compiles_bind_param)
    compiles(Interval, "bigquery")(_compile_interval)
    compiles(CreateIndex, "bigquery")(_compile_create_index)
    compiles(sqlalchemy.sql.functions.now, "bigquery")(_compile_func_now)
    compiles(BinaryExpression, "bigquery")(_compiles_binary_expression)
    compiles(Cast, "bigquery")(_compile_cast)
    compiles(CreateTable, "bigquery")(_compile_create_table)
    compiles(DropTable, "bigquery")(_compile_drop_table)
    compiles(DateTime, "bigquery")(_compile_datetime)
    compiles(STRUCT, "bigquery")(_compile_struct)
    _patch_bigquery_dbapi()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/artifacts/v1/export.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.artifacts.v1 import cdc_pb2 as chalk_dot_artifacts_dot_v1_dot_cdc__pb2
from chalk._gen.chalk.artifacts.v1 import chart_pb2 as chalk_dot_artifacts_dot_v1_dot_chart__pb2
from chalk._gen.chalk.artifacts.v1 import cron_query_pb2 as chalk_dot_artifacts_dot_v1_dot_cron__query__pb2
from chalk._gen.chalk.common.v1 import chalk_error_pb2 as chalk_dot_common_dot_v1_dot_chalk__error__pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as chalk_dot_graph_dot_v1_dot_graph__pb2
from chalk._gen.chalk.lsp.v1 import lsp_pb2 as chalk_dot_lsp_dot_v1_dot_lsp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1f\x63halk/artifacts/v1/export.proto\x12\x12\x63halk.artifacts.v1\x1a\x1c\x63halk/artifacts/v1/cdc.proto\x1a\x1e\x63halk/artifacts/v1/chart.proto\x1a#chalk/artifacts/v1/cron_query.proto\x1a!chalk/common/v1/chalk_error.proto\x1a\x1a\x63halk/graph/v1/graph.proto\x1a\x16\x63halk/lsp/v1/lsp.proto"\xb0\x02\n\x13\x45nvironmentSettings\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x1d\n\x07runtime\x18\x02 \x01(\tH\x00R\x07runtime\x88\x01\x01\x12\'\n\x0crequirements\x18\x03 \x01(\tH\x01R\x0crequirements\x88\x01\x01\x12#\n\ndockerfile\x18\x04 \x01(\tH\x02R\ndockerfile\x88\x01\x01\x12+\n\x11requires_packages\x18\x05 \x03(\tR\x10requiresPackages\x12.\n\x10platform_version\x18\x06 \x01(\tH\x03R\x0fplatformVersion\x88\x01\x01\x42\n\n\x08_runtimeB\x0f\n\r_requirementsB\r\n\x0b_dockerfileB\x13\n\x11_platform_version"\xc0\x01\n\x0fProjectSettings\x12\x18\n\x07project\x18\x01 \x01(\tR\x07project\x12K\n\x0c\x65nvironments\x18\x02 \x03(\x0b\x32\'.chalk.artifacts.v1.EnvironmentSettingsR\x0c\x65nvironments\x12\x46\n\nvalidation\x18\x03 \x01(\x0b\x32&.chalk.artifacts.v1.ValidationSettingsR\nvalidation"@\n\x10MetadataSettings\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x18\n\x07missing\x18\x02 \x01(\tR\x07missing"S\n\x0f\x46\x65\x61tureSettings\x12@\n\x08metadata\x18\x01 \x03(\x0b\x32$.chalk.artifacts.v1.MetadataSettingsR\x08metadata"T\n\x10ResolverSettings\x12@\n\x08metadata\x18\x01 \x03(\x0b\x32$.chalk.artifacts.v1.MetadataSettingsR\x08metadata"\x95\x01\n\x12ValidationSettings\x12=\n\x07\x66\x65\x61ture\x18\x01 \x01(\x0b\x32#.chalk.artifacts.v1.FeatureSettingsR\x07\x66\x65\x61ture\x12@\n\x08resolver\x18\x02 \x01(\x0b\x32$.chalk.artifacts.v1.ResolverSettingsR\x08resolver"a\n\x0c\x46\x61iledImport\x12\x1b\n\tfile_name\x18\x01 \x01(\tR\x08\x66ileName\x12\x16\n\x06module\x18\x02 \x01(\tR\x06module\x12\x1c\n\ttraceback\x18\x03 \x01(\tR\ttraceback"O\n\x0b\x43halkpyInfo\x12\x18\n\x07version\x18\x01 \x01(\tR\x07version\x12\x1b\n\x06python\x18\x02 \x01(\tH\x00R\x06python\x88\x01\x01\x42\t\n\x07_python"\x8c\x01\n\rValidationLog\x12\x16\n\x06header\x18\x01 \x01(\tR\x06header\x12\x1c\n\tsubheader\x18\x02 \x01(\tR\tsubheader\x12\x45\n\x08severity\x18\x03 \x01(\x0e\x32).chalk.artifacts.v1.ValidationLogSeverityR\x08severity"\xb5\x04\n\x06\x45xport\x12+\n\x05graph\x18\x01 \x01(\x0b\x32\x15.chalk.graph.v1.GraphR\x05graph\x12\x33\n\x05\x63rons\x18\x02 \x03(\x0b\x32\x1d.chalk.artifacts.v1.CronQueryR\x05\x63rons\x12\x31\n\x06\x63harts\x18\x03 \x03(\x0b\x32\x19.chalk.artifacts.v1.ChartR\x06\x63harts\x12>\n\x0b\x63\x64\x63_sources\x18\x04 \x03(\x0b\x32\x1d.chalk.artifacts.v1.CDCSourceR\ncdcSources\x12;\n\x06\x63onfig\x18\x05 \x01(\x0b\x32#.chalk.artifacts.v1.ProjectSettingsR\x06\x63onfig\x12\x39\n\x07\x63halkpy\x18\x06 \x01(\x0b\x32\x1f.chalk.artifacts.v1.ChalkpyInfoR\x07\x63halkpy\x12\x38\n\x06\x66\x61iled\x18\x07 \x03(\x0b\x32 .chalk.artifacts.v1.FailedImportR\x06\x66\x61iled\x12\x35\n\x04logs\x18\x08 \x03(\x0b\x32!.chalk.artifacts.v1.ValidationLogR\x04logs\x12#\n\x03lsp\x18\t \x01(\x0b\x32\x11.chalk.lsp.v1.LSPR\x03lsp\x12H\n\x11\x63onversion_errors\x18\n \x03(\x0b\x32\x1b.chalk.common.v1.ChalkErrorR\x10\x63onversionErrors*\xaa\x01\n\x15ValidationLogSeverity\x12\'\n#VALIDATION_LOG_SEVERITY_UNSPECIFIED\x10\x00\x12 \n\x1cVALIDATION_LOG_SEVERITY_INFO\x10\x04\x12#\n\x1fVALIDATION_LOG_SEVERITY_WARNING\x10\x08\x12!\n\x1dVALIDATION_LOG_SEVERITY_ERROR\x10\x0c\x42\x8f\x01\n\x16\x63om.chalk.artifacts.v1B\x0b\x45xportProtoP\x01\xa2\x02\x03\x43\x41X\xaa\x02\x12\x43halk.Artifacts.V1\xca\x02\x12\x43halk\\Artifacts\\V1\xe2\x02\x1e\x43halk\\Artifacts\\V1\\GPBMetadata\xea\x02\x14\x43halk::Artifacts::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.artifacts.v1.export_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\026com.chalk.artifacts.v1B\013ExportProtoP\001\242\002\003CAX\252\002\022Chalk.Artifacts.V1\312\002\022Chalk\\Artifacts\\V1\342\002\036Chalk\\Artifacts\\V1\\GPBMetadata\352\002\024Chalk::Artifacts::V1"
    _globals["_VALIDATIONLOGSEVERITY"]._serialized_start = 2024
    _globals["_VALIDATIONLOGSEVERITY"]._serialized_end = 2194
    _globals["_ENVIRONMENTSETTINGS"]._serialized_start = 242
    _globals["_ENVIRONMENTSETTINGS"]._serialized_end = 546
    _globals["_PROJECTSETTINGS"]._serialized_start = 549
    _globals["_PROJECTSETTINGS"]._serialized_end = 741
    _globals["_METADATASETTINGS"]._serialized_start = 743
    _globals["_METADATASETTINGS"]._serialized_end = 807
    _globals["_FEATURESETTINGS"]._serialized_start = 809
    _globals["_FEATURESETTINGS"]._serialized_end = 892
    _globals["_RESOLVERSETTINGS"]._serialized_start = 894
    _globals["_RESOLVERSETTINGS"]._serialized_end = 978
    _globals["_VALIDATIONSETTINGS"]._serialized_start = 981
    _globals["_VALIDATIONSETTINGS"]._serialized_end = 1130
    _globals["_FAILEDIMPORT"]._serialized_start = 1132
    _globals["_FAILEDIMPORT"]._serialized_end = 1229
    _globals["_CHALKPYINFO"]._serialized_start = 1231
    _globals["_CHALKPYINFO"]._serialized_end = 1310
    _globals["_VALIDATIONLOG"]._serialized_start = 1313
    _globals["_VALIDATIONLOG"]._serialized_end = 1453
    _globals["_EXPORT"]._serialized_start = 1456
    _globals["_EXPORT"]._serialized_end = 2021
# @@protoc_insertion_point(module_scope)



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class LogEntry(_message.Message):
    __slots__ = ("id", "severity", "timestamp", "message")
    ID_FIELD_NUMBER: _ClassVar[int]
    SEVERITY_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    id: str
    severity: str
    timestamp: _timestamp_pb2.Timestamp
    message: str
    def __init__(
        self,
        id: _Optional[str] = ...,
        severity: _Optional[str] = ...,
        timestamp: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        message: _Optional[str] = ...,
    ) -> None: ...

class SearchLogEntriesPageToken(_message.Message):
    __slots__ = ("next_page_token",)
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    next_page_token: str
    def __init__(self, next_page_token: _Optional[str] = ...) -> None: ...

class SearchLogEntriesRequest(_message.Message):
    __slots__ = ("query", "page_token")
    QUERY_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    query: str
    page_token: SearchLogEntriesPageToken
    def __init__(
        self, query: _Optional[str] = ..., page_token: _Optional[_Union[SearchLogEntriesPageToken, _Mapping]] = ...
    ) -> None: ...

class SearchLogEntriesResponse(_message.Message):
    __slots__ = ("log_entries", "next_page_token")
    LOG_ENTRIES_FIELD_NUMBER: _ClassVar[int]
    NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    log_entries: _containers.RepeatedCompositeFieldContainer[LogEntry]
    next_page_token: SearchLogEntriesPageToken
    def __init__(
        self,
        log_entries: _Optional[_Iterable[_Union[LogEntry, _Mapping]]] = ...,
        next_page_token: _Optional[_Union[SearchLogEntriesPageToken, _Mapping]] = ...,
    ) -> None: ...



"""
Adds thread-local context to a Python logger. Taken from neocrym/log-with-context
"""

from __future__ import annotations

import collections.abc
import contextlib
import contextvars
import functools
import logging
import logging.config
import os
import sys
import threading
import types
from enum import Enum
from typing import Any, Dict, Mapping, Optional
from weakref import WeakKeyDictionary

from chalk.utils._ddtrace_version import can_use_ddtrace
from chalk.utils.missing_dependency import missing_dependency_exception

_LOGGING_CONTEXT: contextvars.ContextVar[Mapping[str, Any]] = contextvars.ContextVar("_LOGGING_CONTEXT", default={})


def _recursive_merge(a: Mapping[str, Any], b: Mapping[str, Any]):
    ans: Dict[str, Any] = {**a}
    for k, v in b.items():
        if k not in ans or not (isinstance(ans[k], collections.abc.Mapping) and isinstance(v, collections.abc.Mapping)):
            ans[k] = v
            continue
        ans[k] = _recursive_merge(ans[k], v)
    return ans


def get_logging_context() -> Mapping[str, Any]:
    """
    Retrieve the log context for the current python context.
    This initializes the thread-local variable if necessary.
    """
    return _LOGGING_CONTEXT.get({})


class LogWithContextFilter(logging.Filter):
    """Filter to append the ``extras`` onto the LogRecord."""

    def filter(self, record: logging.LogRecord) -> bool:
        items = get_logging_context().items()
        for k, v in items:
            if not hasattr(record, k):
                setattr(record, k, v)
        return True


filtered_loggers = WeakKeyDictionary()


def get_logger(name: Optional[str]):
    logger = logging.getLogger(name)
    if logger not in filtered_loggers:
        logger.addFilter(LogWithContextFilter())
        filtered_loggers[logger] = True
    return logger


# Backwards compatibility
Logger = get_logger

_logger = get_logger(__name__)


@contextlib.contextmanager
def add_logging_context(*, _merge: bool = True, **log_context: Any):
    """A context manager to push and pop `extra` dictionary keys.

    Parameters
    ----------
    _merge
        Whether to merge the new context with the existing log context.
    extra
        Contextual information to add to the log record
    """
    if _merge:
        log_context = _recursive_merge(
            _LOGGING_CONTEXT.get({}),
            log_context,
        )
    token = _LOGGING_CONTEXT.set(log_context)
    try:
        yield
    finally:
        _LOGGING_CONTEXT.reset(token)


_PUBLIC_LOGGERS_WITH_FILTER: set[str] = set()


class _PublicLoggingFilter(logging.Filter):
    def filter(self, record: logging.LogRecord):
        record.is_public = True
        return True


def get_public_logger(name: str):
    name += ".public"
    logger = get_logger(name)
    if name not in _PUBLIC_LOGGERS_WITH_FILTER:
        _PUBLIC_LOGGERS_WITH_FILTER.add(name)
        logger.addFilter(_PublicLoggingFilter())
    return logger


OPERATION_KEY = "operation"
LABELS_KEY = "labels"
_OPERATION_ID_KEY = "operation_id"
_OPERATION_PRODUCER = "operation_producer"
_OPERATION_IS_FIRST = "operation_is_first"
_OPERATION_IS_LAST = "operation_is_last"


def get_operation_log_context(
    operation_id: str,
    operation_kind: str | Enum,
    resolver_fqn: str | None = None,
    query_name: str | None = None,
    correlation_id: str | None = None,
):
    """
    :param operation_id: internal correlation id
    :param operation_kind: ...
    :param resolver_fqn: ...
    :param query_name: customer-provided query name
    :param correlation_id: customer-provided correlation id
    :return:
    """
    if isinstance(operation_kind, Enum):
        operation_kind = operation_kind.value
    ctx: dict[str, Any] = {
        _OPERATION_ID_KEY: operation_id,
        _OPERATION_PRODUCER: operation_kind,
        LABELS_KEY: {},
    }
    if resolver_fqn is not None:
        ctx[LABELS_KEY]["resolver_fqn"] = resolver_fqn
    if query_name is not None:
        ctx[LABELS_KEY]["query_name"] = query_name
    if correlation_id is not None:
        ctx["correlation_id"] = correlation_id
    return ctx


IS_FIRST_OPERATION_CTX = {_OPERATION_IS_FIRST: True}
IS_LAST_OPERATION_CTX = {_OPERATION_IS_LAST: True}


@functools.lru_cache(None)
def get_json_logging_formatter() -> logging.Formatter:
    try:
        from pythonjsonlogger import jsonlogger
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")

    if can_use_ddtrace:
        import ddtrace
    else:
        ddtrace = None

    class ChalkJsonFormatter(jsonlogger.JsonFormatter):
        def __init__(self):
            super().__init__(
                reserved_attrs=[
                    "args",
                    "msg",
                    "levelno",
                    "filename",
                    "msecs",
                    "relativeCreated",
                    "module",
                    "exc_info",
                    "stack_info",
                ],
            )
            self._dd_entity_id = os.getenv("DD_ENTITY_ID", None)
            try:
                import google.cloud.client

                client = google.cloud.client.ClientWithProject()
            except:
                # Likely no credentials available
                self._gcp_project_name = None
            else:
                self._gcp_project_name = client.project

        def process_log_record(self, log_record: Dict[str, Any]):
            # We want to duplicate some fields, so it will show up nicely for google structured logging
            log_record.update(get_logging_context())
            log_record["severity"] = log_record["levelname"]
            if "exc_info" in log_record:
                log_record["message"] = log_record.get("message", "") + "\n" + log_record["exc_info"]
                del log_record["exc_info"]
            source_location = {
                "file": log_record["pathname"],
                "line": str(log_record["lineno"]),
            }
            if len(log_record.get("message", "")) > 1024 * 250:
                # Datadog truncates log messages at 256KB
                # see https://docs.datadoghq.com/agent/logs/log_transport/?tab=https#https-transport
                log_record["message"] = log_record["message"] + "... (log message truncated due to size)"
            if log_record.get("funcName") is not None:
                source_location["function"] = log_record["funcName"]
            log_record["logging.googleapis.com/sourceLocation"] = source_location
            if self._dd_entity_id is not None:
                log_record["dd.entity_id"] = self._dd_entity_id
            if ddtrace is not None:
                if ddtrace.config.env is not None:
                    log_record["dd.env"] = ddtrace.config.env
                if ddtrace.config.service is not None:
                    log_record["dd.service"] = ddtrace.config.service
                if ddtrace.config.version is not None:
                    log_record["dd.version"] = ddtrace.config.version

                if (current_trace_context := ddtrace.tracer.current_trace_context()) is not None:
                    if current_trace_context.trace_id is not None:
                        log_record["dd.trace_id"] = current_trace_context.trace_id
                        if self._gcp_project_name is not None:
                            log_record[
                                "logging.googleapis.com/trace"
                            ] = f"projects/{self._gcp_project_name}/traces/{current_trace_context.trace_id}"
                    if current_trace_context.span_id is not None:
                        log_record["dd.span_id"] = current_trace_context.span_id
                        log_record["logging.googleapis.com/spanId"] = current_trace_context.span_id
                    log_record["logging.googleapis.com/trace_sampled"] = (
                        current_trace_context.sampling_priority is not None
                        and current_trace_context.sampling_priority > 0
                    )
            operation = {}
            if (operation_id := log_record.get(_OPERATION_ID_KEY)) is not None:
                operation["id"] = operation_id
            if (operation_producer := log_record.get(_OPERATION_PRODUCER)) is not None:
                operation["producer"] = operation_producer
            if (operation_is_first := log_record.get(_OPERATION_IS_FIRST, None)) is not None:
                operation["first"] = operation_is_first
            if (operation_is_last := log_record.get(_OPERATION_IS_LAST, None)) is not None:
                operation["last"] = operation_is_last
            if operation:
                log_record["logging.googleapis.com/operation"] = operation
            log_record["logging.googleapis.com/labels"] = log_record.get(LABELS_KEY, {})
            log_record["timestampSeconds"] = int(log_record["created"])
            log_record["timestampNanos"] = int((log_record["created"] - int(log_record["created"])) * 1e9)
            log_record = {k: v for (k, v) in log_record.items() if v is not None}
            return log_record

    return ChalkJsonFormatter()


def _threading_excepthook(args: threading.ExceptHookArgs):
    typ = args.exc_type
    value = args.exc_value
    traceback = args.exc_traceback
    thread = args.thread
    if typ is not None and value is not None:  # pyright: ignore[reportUnnecessaryComparison]
        try:
            _logger.error(
                f"Unhandled root exception in thread{'' if thread is None else f' {thread.ident} ({thread.getName()})'}",
                exc_info=(typ, value, traceback),
            )
        except:
            # Call the original excepthook in case of error
            # This function exists even though pyright doesn't know about it
            # See https://docs.python.org/3/library/threading.html#threading.__excepthook__
            threading.__excepthook__(args)  # pyright: ignore
    return True


def _sys_excepthook(
    typ: type[BaseException] | None = None,
    value: BaseException | None = None,
    traceback: types.TracebackType | None = None,
):
    if typ is not None and value is not None:
        try:
            _logger.error(
                "Unhandled root exception",
                exc_info=(typ, value, traceback),
            )
        except:
            # Call the original excepthook in case of error
            sys.__excepthook__(typ, value, traceback)
    return True


def configure_logging(logging_config: dict[str, Any]):
    logging.config.dictConfig(logging_config)
    logging.captureWarnings(True)
    _logger.info(f"Configured logging with config: `{logging_config}`")
    # If we are using structured logging, then route unhandled exceptions through the logger so they will be json-formatted
    sys.excepthook = _sys_excepthook
    threading.excepthook = _threading_excepthook



# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: ignore-errors
# NOTE: This file was copied from https://raw.githubusercontent.com/sqlalchemy/sqlalchemy/rel_2_0_13/lib/sqlalchemy/dialects/postgresql/psycopg.py.
# Chalk modified it to be compatible with sqlalchemy 1.4

r"""
.. dialect:: postgresql+psycopg
    :name: psycopg (a.k.a. psycopg 3)
    :dbapi: psycopg
    :connectstring: postgresql+psycopg://user:password@host:port/dbname[?key=value&key=value...]
    :url: https://pypi.org/project/psycopg/

``psycopg`` is the package and module name for version 3 of the ``psycopg``
database driver, formerly known as ``psycopg2``.  This driver is different
enough from its ``psycopg2`` predecessor that SQLAlchemy supports it
via a totally separate dialect; support for ``psycopg2`` is expected to remain
for as long as that package continues to function for modern Python versions,
and also remains the default dialect for the ``postgresql://`` dialect
series.

The SQLAlchemy ``psycopg`` dialect provides both a sync and an async
implementation under the same dialect name. The proper version is
selected depending on how the engine is created:

* calling :func:`_sa.create_engine` with ``postgresql+psycopg://...`` will
  automatically select the sync version, e.g.::

    from sqlalchemy import create_engine
    sync_engine = create_engine("postgresql+psycopg://scott:tiger@localhost/test")

* calling :func:`_asyncio.create_async_engine` with
  ``postgresql+psycopg://...`` will automatically select the async version,
  e.g.::

    from sqlalchemy.ext.asyncio import create_async_engine
    asyncio_engine = create_async_engine("postgresql+psycopg://scott:tiger@localhost/test")

The asyncio version of the dialect may also be specified explicitly using the
``psycopg_async`` suffix, as::

    from sqlalchemy.ext.asyncio import create_async_engine
    asyncio_engine = create_async_engine("postgresql+psycopg_async://scott:tiger@localhost/test")

The ``psycopg`` dialect has the same API features as that of ``psycopg2``,
with the exception of the "fast executemany" helpers.   The "fast executemany"
helpers are expected to be generalized and ported to ``psycopg`` before the final
release of SQLAlchemy 2.0, however.


.. seealso::

    :ref:`postgresql_psycopg2` - The SQLAlchemy ``psycopg``
    dialect shares most of its behavior with the ``psycopg2`` dialect.
    Further documentation is available there.

"""  # noqa
from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, cast

from sqlalchemy import pool, util
from sqlalchemy.dialects.postgresql import ranges
from sqlalchemy.dialects.postgresql.base import INTERVAL, PGCompiler, PGIdentifierPreparer

# from sqlalchemy.dialects.postgresql.base import REGCONFIG
from sqlalchemy.dialects.postgresql.json import JSON, JSONB, JSONPathType
from sqlalchemy.engine import AdaptedConnection
from sqlalchemy.sql import sqltypes
from sqlalchemy.util.concurrency import await_fallback, await_only

from chalk.sql._internal.integrations.psycopg3.psycopg_common import (
    _PGDialect_common_psycopg,
    _PGExecutionContext_common_psycopg,
)

if TYPE_CHECKING:
    from typing import Iterable

logger = logging.getLogger("sqlalchemy.dialects.postgresql")


class _PGString(sqltypes.String):
    render_bind_cast = True


# class _PGREGCONFIG(REGCONFIG):
#     render_bind_cast = True


class _PGJSON(JSON):
    render_bind_cast = True

    def bind_processor(self, dialect):
        return self._make_bind_processor(None, dialect._psycopg_Json)

    def result_processor(self, dialect, coltype):
        return None


class _PGJSONB(JSONB):
    render_bind_cast = True

    def bind_processor(self, dialect):
        return self._make_bind_processor(None, dialect._psycopg_Jsonb)

    def result_processor(self, dialect, coltype):
        return None


class _PGJSONIntIndexType(sqltypes.JSON.JSONIntIndexType):
    __visit_name__ = "json_int_index"

    render_bind_cast = True


class _PGJSONStrIndexType(sqltypes.JSON.JSONStrIndexType):
    __visit_name__ = "json_str_index"

    render_bind_cast = True


class _PGJSONPathType(JSONPathType):
    pass


class _PGInterval(INTERVAL):
    render_bind_cast = True


class _PGTimeStamp(sqltypes.DateTime):
    render_bind_cast = True


class _PGDate(sqltypes.Date):
    render_bind_cast = True


class _PGTime(sqltypes.Time):
    render_bind_cast = True


class _PGInteger(sqltypes.Integer):
    render_bind_cast = True


class _PGSmallInteger(sqltypes.SmallInteger):
    render_bind_cast = True


class _PGNullType(sqltypes.NullType):
    render_bind_cast = True


class _PGBigInteger(sqltypes.BigInteger):
    render_bind_cast = True


class _PGBoolean(sqltypes.Boolean):
    render_bind_cast = True


# class _PsycopgRange(ranges.AbstractRangeImpl):
#     def bind_processor(self, dialect):
#         psycopg_Range = cast(PGDialect_psycopg, dialect)._psycopg_Range

#         def to_range(value):
#             if isinstance(value, ranges.Range):
#                 value = psycopg_Range(
#                     value.lower, value.upper, value.bounds, value.empty
#                 )
#             return value

#         return to_range

#     def result_processor(self, dialect, coltype):
#         def to_range(value):
#             if value is not None:
#                 value = ranges.Range(
#                     value._lower,
#                     value._upper,
#                     bounds=value._bounds if value._bounds else "[)",
#                     empty=not value._bounds,
#                 )
#             return value

#         return to_range


# class _PsycopgMultiRange(ranges.AbstractMultiRangeImpl):
#     def bind_processor(self, dialect):
#         psycopg_Range = cast(PGDialect_psycopg, dialect)._psycopg_Range
#         psycopg_Multirange = cast(
#             PGDialect_psycopg, dialect
#         )._psycopg_Multirange

#         NoneType = type(None)

#         def to_range(value):
#             if isinstance(value, (str, NoneType, psycopg_Multirange)):
#                 return value

#             return psycopg_Multirange(
#                 [
#                     psycopg_Range(
#                         element.lower,
#                         element.upper,
#                         element.bounds,
#                         element.empty,
#                     )
#                     for element in cast("Iterable[ranges.Range]", value)
#                 ]
#             )

#         return to_range

#     def result_processor(self, dialect, coltype):
#         def to_range(value):
#             if value is not None:
#                 value = [
#                     ranges.Range(
#                         elem._lower,
#                         elem._upper,
#                         bounds=elem._bounds if elem._bounds else "[)",
#                         empty=not elem._bounds,
#                     )
#                     for elem in value
#                 ]

#             return value

#         return to_range


class PGExecutionContext_psycopg(_PGExecutionContext_common_psycopg):
    pass


class PGCompiler_psycopg(PGCompiler):
    pass


class PGIdentifierPreparer_psycopg(PGIdentifierPreparer):
    pass


def _log_notices(diagnostic):
    logger.info("%s: %s", diagnostic.severity, diagnostic.message_primary)


class PGDialect_psycopg(_PGDialect_common_psycopg):
    driver = "psycopg"

    supports_statement_cache = True
    supports_server_side_cursors = True
    default_paramstyle = "pyformat"
    supports_sane_multi_rowcount = True

    execution_ctx_cls = PGExecutionContext_psycopg
    statement_compiler = PGCompiler_psycopg
    preparer = PGIdentifierPreparer_psycopg
    psycopg_version = (0, 0)

    _has_native_hstore = True
    _psycopg_adapters_map = None

    colspecs = util.update_copy(
        _PGDialect_common_psycopg.colspecs,
        {
            sqltypes.String: _PGString,
            # REGCONFIG: _PGREGCONFIG,
            JSON: _PGJSON,
            sqltypes.JSON: _PGJSON,
            JSONB: _PGJSONB,
            sqltypes.JSON.JSONPathType: _PGJSONPathType,
            sqltypes.JSON.JSONIntIndexType: _PGJSONIntIndexType,
            sqltypes.JSON.JSONStrIndexType: _PGJSONStrIndexType,
            sqltypes.Interval: _PGInterval,
            INTERVAL: _PGInterval,
            sqltypes.Date: _PGDate,
            sqltypes.DateTime: _PGTimeStamp,
            sqltypes.Time: _PGTime,
            sqltypes.Integer: _PGInteger,
            sqltypes.SmallInteger: _PGSmallInteger,
            sqltypes.BigInteger: _PGBigInteger,
            # ranges.AbstractRange: _PsycopgRange,
            # ranges.AbstractMultiRange: _PsycopgMultiRange,
        },
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.dbapi:
            m = re.match(r"(\d+)\.(\d+)(?:\.(\d+))?", self.dbapi.__version__)
            if m:
                self.psycopg_version = tuple(int(x) for x in m.group(1, 2, 3) if x is not None)

            if self.psycopg_version < (3, 0, 2):
                raise ImportError("psycopg version 3.0.2 or higher is required.")

            from psycopg.adapt import AdaptersMap

            self._psycopg_adapters_map = adapters_map = AdaptersMap(self.dbapi.adapters)

            if self._json_deserializer:
                from psycopg.types.json import set_json_loads

                set_json_loads(self._json_deserializer, adapters_map)

            if self._json_serializer:
                from psycopg.types.json import set_json_dumps

                set_json_dumps(self._json_serializer, adapters_map)

    def create_connect_args(self, url):
        # see https://github.com/psycopg/psycopg/issues/83
        cargs, cparams = super().create_connect_args(url)

        if self._psycopg_adapters_map:
            cparams["context"] = self._psycopg_adapters_map
        if self.client_encoding is not None:
            cparams["client_encoding"] = self.client_encoding
        return cargs, cparams

    def _type_info_fetch(self, connection, name):
        from psycopg.types import TypeInfo

        return TypeInfo.fetch(connection.connection.driver_connection, name)

    def initialize(self, connection):
        super().initialize(connection)

        # PGDialect.initialize() checks server version for <= 8.2 and sets
        # this flag to False if so
        if not self.full_returning:
            self.insert_executemany_returning = False

        # HSTORE can't be registered until we have a connection so that
        # we can look up its OID, so we set up this adapter in
        # initialize()
        if self.use_native_hstore:
            info = self._type_info_fetch(connection, "hstore")
            self._has_native_hstore = info is not None
            if self._has_native_hstore:
                from psycopg.types.hstore import register_hstore

                # register the adapter for connections made subsequent to
                # this one
                register_hstore(info, self._psycopg_adapters_map)

                # register the adapter for this connection
                register_hstore(info, connection.connection)

    @classmethod
    def dbapi(cls):
        import psycopg

        return psycopg

    @classmethod
    def get_async_dialect_cls(cls, url):
        return PGDialectAsync_psycopg

    @util.memoized_property
    def _isolation_lookup(self):
        return {
            "READ COMMITTED": self.dbapi.IsolationLevel.READ_COMMITTED,
            "READ UNCOMMITTED": self.dbapi.IsolationLevel.READ_UNCOMMITTED,
            "REPEATABLE READ": self.dbapi.IsolationLevel.REPEATABLE_READ,
            "SERIALIZABLE": self.dbapi.IsolationLevel.SERIALIZABLE,
        }

    @util.memoized_property
    def _psycopg_Json(self):
        from psycopg.types import json

        return json.Json

    @util.memoized_property
    def _psycopg_Jsonb(self):
        from psycopg.types import json

        return json.Jsonb

    @util.memoized_property
    def _psycopg_TransactionStatus(self):
        from psycopg.pq import TransactionStatus

        return TransactionStatus

    @util.memoized_property
    def _psycopg_Range(self):
        from psycopg.types.range import Range

        return Range

    @util.memoized_property
    def _psycopg_Multirange(self):
        from psycopg.types.multirange import Multirange

        return Multirange

    def _do_isolation_level(self, connection, autocommit, isolation_level):
        connection.autocommit = autocommit
        connection.isolation_level = isolation_level

    def get_isolation_level(self, dbapi_connection):
        status_before = dbapi_connection.info.transaction_status
        value = super().get_isolation_level(dbapi_connection)

        # don't rely on psycopg providing enum symbols, compare with
        # eq/ne
        if status_before == self._psycopg_TransactionStatus.IDLE:
            dbapi_connection.rollback()
        return value

    def set_isolation_level(self, dbapi_connection, level):
        if level == "AUTOCOMMIT":
            self._do_isolation_level(dbapi_connection, autocommit=True, isolation_level=None)
        else:
            self._do_isolation_level(
                dbapi_connection,
                autocommit=False,
                isolation_level=self._isolation_lookup[level],
            )

    def set_readonly(self, connection, value):
        connection.read_only = value

    def get_readonly(self, connection):
        return connection.read_only

    def on_connect(self):
        def notices(conn):
            conn.add_notice_handler(_log_notices)

        fns = [notices]

        if self.isolation_level is not None:

            def on_connect(conn):
                self.set_isolation_level(conn, self.isolation_level)

            fns.append(on_connect)

        # fns always has the notices function
        def on_connect(conn):
            for fn in fns:
                fn(conn)

        return on_connect

    def is_disconnect(self, e, connection, cursor):
        if isinstance(e, self.dbapi.Error) and connection is not None:
            if connection.closed or connection.broken:
                return True
        return False

    def _do_prepared_twophase(self, connection, command, recover=False):
        dbapi_conn = connection.connection.dbapi_connection
        if (
            recover
            # don't rely on psycopg providing enum symbols, compare with
            # eq/ne
            or dbapi_conn.info.transaction_status != self._psycopg_TransactionStatus.IDLE
        ):
            dbapi_conn.rollback()
        before_autocommit = dbapi_conn.autocommit
        try:
            if not before_autocommit:
                self._do_autocommit(dbapi_conn, True)
            dbapi_conn.execute(command)
        finally:
            if not before_autocommit:
                self._do_autocommit(dbapi_conn, before_autocommit)

    def do_rollback_twophase(self, connection, xid, is_prepared=True, recover=False):
        if is_prepared:
            self._do_prepared_twophase(connection, f"ROLLBACK PREPARED '{xid}'", recover=recover)
        else:
            self.do_rollback(connection.connection)

    def do_commit_twophase(self, connection, xid, is_prepared=True, recover=False):
        if is_prepared:
            self._do_prepared_twophase(connection, f"COMMIT PREPARED '{xid}'", recover=recover)
        else:
            self.do_commit(connection.connection)

    @util.memoized_property
    def _dialect_specific_select_one(self):
        return ";"


class AsyncAdapt_psycopg_cursor:
    __slots__ = ("_cursor", "await_", "_rows")

    _psycopg_ExecStatus = None

    def __init__(self, cursor, await_) -> None:
        self._cursor = cursor
        self.await_ = await_
        self._rows = []

    def __getattr__(self, name):
        return getattr(self._cursor, name)

    @property
    def arraysize(self):
        return self._cursor.arraysize

    @arraysize.setter
    def arraysize(self, value):
        self._cursor.arraysize = value

    def close(self):
        self._rows.clear()
        # Normal cursor just call _close() in a non-sync way.
        self._cursor._close()

    def execute(self, query, params=None, **kw):
        result = self.await_(self._cursor.execute(query, params, **kw))
        # sqlalchemy result is not async, so need to pull all rows here
        res = self._cursor.pgresult

        # don't rely on psycopg providing enum symbols, compare with
        # eq/ne
        if res and res.status == self._psycopg_ExecStatus.TUPLES_OK:
            rows = self.await_(self._cursor.fetchall())
            if not isinstance(rows, list):
                self._rows = list(rows)
            else:
                self._rows = rows
        return result

    def executemany(self, query, params_seq):
        return self.await_(self._cursor.executemany(query, params_seq))

    def __iter__(self):
        # TODO: try to avoid pop(0) on a list
        while self._rows:
            yield self._rows.pop(0)

    def fetchone(self):
        if self._rows:
            # TODO: try to avoid pop(0) on a list
            return self._rows.pop(0)
        else:
            return None

    def fetchmany(self, size=None):
        if size is None:
            size = self._cursor.arraysize

        retval = self._rows[0:size]
        self._rows = self._rows[size:]
        return retval

    def fetchall(self):
        retval = self._rows
        self._rows = []
        return retval


class AsyncAdapt_psycopg_ss_cursor(AsyncAdapt_psycopg_cursor):
    def execute(self, query, params=None, **kw):
        self.await_(self._cursor.execute(query, params, **kw))
        return self

    def close(self):
        self.await_(self._cursor.close())

    def fetchone(self):
        return self.await_(self._cursor.fetchone())

    def fetchmany(self, size=0):
        return self.await_(self._cursor.fetchmany(size))

    def fetchall(self):
        return self.await_(self._cursor.fetchall())

    def __iter__(self):
        iterator = self._cursor.__aiter__()
        while True:
            try:
                yield self.await_(iterator.__anext__())
            except StopAsyncIteration:
                break


class AsyncAdapt_psycopg_connection(AdaptedConnection):
    __slots__ = ()
    await_ = staticmethod(await_only)

    def __init__(self, connection) -> None:
        self._connection = connection

    def __getattr__(self, name):
        return getattr(self._connection, name)

    def execute(self, query, params=None, **kw):
        cursor = self.await_(self._connection.execute(query, params, **kw))
        return AsyncAdapt_psycopg_cursor(cursor, self.await_)

    def cursor(self, *args, **kw):
        cursor = self._connection.cursor(*args, **kw)
        if hasattr(cursor, "name"):
            return AsyncAdapt_psycopg_ss_cursor(cursor, self.await_)
        else:
            return AsyncAdapt_psycopg_cursor(cursor, self.await_)

    def commit(self):
        self.await_(self._connection.commit())

    def rollback(self):
        self.await_(self._connection.rollback())

    def close(self):
        self.await_(self._connection.close())

    @property
    def autocommit(self):
        return self._connection.autocommit

    @autocommit.setter
    def autocommit(self, value):
        self.set_autocommit(value)

    def set_autocommit(self, value):
        self.await_(self._connection.set_autocommit(value))

    def set_isolation_level(self, value):
        self.await_(self._connection.set_isolation_level(value))

    def set_read_only(self, value):
        self.await_(self._connection.set_read_only(value))

    def set_deferrable(self, value):
        self.await_(self._connection.set_deferrable(value))


class AsyncAdaptFallback_psycopg_connection(AsyncAdapt_psycopg_connection):
    __slots__ = ()
    await_ = staticmethod(await_fallback)


class PsycopgAdaptDBAPI:
    def __init__(self, psycopg) -> None:
        self.psycopg = psycopg

        for k, v in self.psycopg.__dict__.items():
            if k != "connect":
                self.__dict__[k] = v

    def connect(self, *arg, **kw):
        async_fallback = kw.pop("async_fallback", False)
        if util.asbool(async_fallback):
            return AsyncAdaptFallback_psycopg_connection(
                await_fallback(self.psycopg.AsyncConnection.connect(*arg, **kw))
            )
        else:
            return AsyncAdapt_psycopg_connection(await_only(self.psycopg.AsyncConnection.connect(*arg, **kw)))


class PGDialectAsync_psycopg(PGDialect_psycopg):
    is_async = True
    supports_statement_cache = True

    @classmethod
    def dbapi(cls):
        import psycopg
        from psycopg.pq import ExecStatus

        AsyncAdapt_psycopg_cursor._psycopg_ExecStatus = ExecStatus

        return PsycopgAdaptDBAPI(psycopg)

    @classmethod
    def get_pool_class(cls, url):

        async_fallback = url.query.get("async_fallback", False)

        if util.asbool(async_fallback):
            return pool.FallbackAsyncAdaptedQueuePool
        else:
            return pool.AsyncAdaptedQueuePool

    def _type_info_fetch(self, connection, name):
        from psycopg.types import TypeInfo

        adapted = connection.connection
        return adapted.await_(TypeInfo.fetch(adapted.driver_connection, name))

    def _do_isolation_level(self, connection, autocommit, isolation_level):
        connection.set_autocommit(autocommit)
        connection.set_isolation_level(isolation_level)

    def _do_autocommit(self, connection, value):
        connection.set_autocommit(value)

    def set_readonly(self, connection, value):
        connection.set_read_only(value)

    def set_deferrable(self, connection, value):
        connection.set_deferrable(value)

    def get_driver_connection(self, connection):
        return connection._connection


dialect = PGDialect_psycopg
dialect_async = PGDialectAsync_psycopg



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/engine/v1/query_server.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.aggregate.v1 import service_pb2 as chalk_dot_aggregate_dot_v1_dot_service__pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as chalk_dot_auth_dot_v1_dot_permissions__pb2
from chalk._gen.chalk.common.v1 import online_query_pb2 as chalk_dot_common_dot_v1_dot_online__query__pb2
from chalk._gen.chalk.common.v1 import upload_features_pb2 as chalk_dot_common_dot_v1_dot_upload__features__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n"chalk/engine/v1/query_server.proto\x12\x0f\x63halk.engine.v1\x1a chalk/aggregate/v1/service.proto\x1a\x1f\x63halk/auth/v1/permissions.proto\x1a"chalk/common/v1/online_query.proto\x1a%chalk/common/v1/upload_features.proto"\x1f\n\x0bPingRequest\x12\x10\n\x03num\x18\x01 \x01(\x05R\x03num" \n\x0cPingResponse\x12\x10\n\x03num\x18\x01 \x01(\x05R\x03num2\xe4\x06\n\x0cQueryService\x12K\n\x04Ping\x12\x1c.chalk.engine.v1.PingRequest\x1a\x1d.chalk.engine.v1.PingResponse"\x06\x90\x02\x01\x80}\x01\x12]\n\x0bOnlineQuery\x12#.chalk.common.v1.OnlineQueryRequest\x1a$.chalk.common.v1.OnlineQueryResponse"\x03\x80}\x03\x12i\n\x0fOnlineQueryBulk\x12\'.chalk.common.v1.OnlineQueryBulkRequest\x1a(.chalk.common.v1.OnlineQueryBulkResponse"\x03\x80}\x03\x12l\n\x10OnlineQueryMulti\x12(.chalk.common.v1.OnlineQueryMultiRequest\x1a).chalk.common.v1.OnlineQueryMultiResponse"\x03\x80}\x03\x12r\n\x12UploadFeaturesBulk\x12*.chalk.common.v1.UploadFeaturesBulkRequest\x1a+.chalk.common.v1.UploadFeaturesBulkResponse"\x03\x80}\x03\x12\x66\n\x0eUploadFeatures\x12&.chalk.common.v1.UploadFeaturesRequest\x1a\'.chalk.common.v1.UploadFeaturesResponse"\x03\x80}\x03\x12\x84\x01\n\x15PlanAggregateBackfill\x12\x30.chalk.aggregate.v1.PlanAggregateBackfillRequest\x1a\x31.chalk.aggregate.v1.PlanAggregateBackfillResponse"\x06\x90\x02\x01\x80}\x0c\x12l\n\rGetAggregates\x12(.chalk.aggregate.v1.GetAggregatesRequest\x1a).chalk.aggregate.v1.GetAggregatesResponse"\x06\x90\x02\x01\x80}\x0b\x42\x85\x01\n\x13\x63om.chalk.engine.v1B\x10QueryServerProtoP\x01\xa2\x02\x03\x43\x45X\xaa\x02\x0f\x43halk.Engine.V1\xca\x02\x0f\x43halk\\Engine\\V1\xe2\x02\x1b\x43halk\\Engine\\V1\\GPBMetadata\xea\x02\x11\x43halk::Engine::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.engine.v1.query_server_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.engine.v1B\020QueryServerProtoP\001\242\002\003CEX\252\002\017Chalk.Engine.V1\312\002\017Chalk\\Engine\\V1\342\002\033Chalk\\Engine\\V1\\GPBMetadata\352\002\021Chalk::Engine::V1"
    _globals["_QUERYSERVICE"].methods_by_name["Ping"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["Ping"]._serialized_options = b"\220\002\001\200}\001"
    _globals["_QUERYSERVICE"].methods_by_name["OnlineQuery"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["OnlineQuery"]._serialized_options = b"\200}\003"
    _globals["_QUERYSERVICE"].methods_by_name["OnlineQueryBulk"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["OnlineQueryBulk"]._serialized_options = b"\200}\003"
    _globals["_QUERYSERVICE"].methods_by_name["OnlineQueryMulti"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["OnlineQueryMulti"]._serialized_options = b"\200}\003"
    _globals["_QUERYSERVICE"].methods_by_name["UploadFeaturesBulk"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["UploadFeaturesBulk"]._serialized_options = b"\200}\003"
    _globals["_QUERYSERVICE"].methods_by_name["UploadFeatures"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["UploadFeatures"]._serialized_options = b"\200}\003"
    _globals["_QUERYSERVICE"].methods_by_name["PlanAggregateBackfill"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["PlanAggregateBackfill"]._serialized_options = b"\220\002\001\200}\014"
    _globals["_QUERYSERVICE"].methods_by_name["GetAggregates"]._options = None
    _globals["_QUERYSERVICE"].methods_by_name["GetAggregates"]._serialized_options = b"\220\002\001\200}\013"
    _globals["_PINGREQUEST"]._serialized_start = 197
    _globals["_PINGREQUEST"]._serialized_end = 228
    _globals["_PINGRESPONSE"]._serialized_start = 230
    _globals["_PINGRESPONSE"]._serialized_end = 262
    _globals["_QUERYSERVICE"]._serialized_start = 265
    _globals["_QUERYSERVICE"]._serialized_end = 1133
# @@protoc_insertion_point(module_scope)



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class DeploymentStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DEPLOYMENT_STATUS_UNSPECIFIED: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_UNKNOWN: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_PENDING: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_QUEUED: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_WORKING: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_SUCCESS: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_FAILURE: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_INTERNAL_ERROR: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_TIMEOUT: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_CANCELLED: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_EXPIRED: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_BOOT_ERRORS: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_AWAITING_SOURCE: _ClassVar[DeploymentStatus]
    DEPLOYMENT_STATUS_DEPLOYING: _ClassVar[DeploymentStatus]

class DeploymentProfilingMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DEPLOYMENT_PROFILING_MODE_UNSPECIFIED: _ClassVar[DeploymentProfilingMode]
    DEPLOYMENT_PROFILING_MODE_NONE: _ClassVar[DeploymentProfilingMode]
    DEPLOYMENT_PROFILING_MODE_O2: _ClassVar[DeploymentProfilingMode]

DEPLOYMENT_STATUS_UNSPECIFIED: DeploymentStatus
DEPLOYMENT_STATUS_UNKNOWN: DeploymentStatus
DEPLOYMENT_STATUS_PENDING: DeploymentStatus
DEPLOYMENT_STATUS_QUEUED: DeploymentStatus
DEPLOYMENT_STATUS_WORKING: DeploymentStatus
DEPLOYMENT_STATUS_SUCCESS: DeploymentStatus
DEPLOYMENT_STATUS_FAILURE: DeploymentStatus
DEPLOYMENT_STATUS_INTERNAL_ERROR: DeploymentStatus
DEPLOYMENT_STATUS_TIMEOUT: DeploymentStatus
DEPLOYMENT_STATUS_CANCELLED: DeploymentStatus
DEPLOYMENT_STATUS_EXPIRED: DeploymentStatus
DEPLOYMENT_STATUS_BOOT_ERRORS: DeploymentStatus
DEPLOYMENT_STATUS_AWAITING_SOURCE: DeploymentStatus
DEPLOYMENT_STATUS_DEPLOYING: DeploymentStatus
DEPLOYMENT_PROFILING_MODE_UNSPECIFIED: DeploymentProfilingMode
DEPLOYMENT_PROFILING_MODE_NONE: DeploymentProfilingMode
DEPLOYMENT_PROFILING_MODE_O2: DeploymentProfilingMode

class InstanceSizing(_message.Message):
    __slots__ = ("min_instances", "max_instances")
    MIN_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    MAX_INSTANCES_FIELD_NUMBER: _ClassVar[int]
    min_instances: int
    max_instances: int
    def __init__(self, min_instances: _Optional[int] = ..., max_instances: _Optional[int] = ...) -> None: ...

class Deployment(_message.Message):
    __slots__ = (
        "id",
        "environment_id",
        "status",
        "deployment_tags",
        "cloud_build_id",
        "triggered_by",
        "requirements_filepath",
        "dockerfile_filepath",
        "runtime",
        "chalkpy_version",
        "raw_dependency_hash",
        "final_dependency_hash",
        "is_preview_deployment",
        "created_at",
        "updated_at",
        "git_commit",
        "git_pr",
        "git_branch",
        "git_author_email",
        "branch",
        "project_settings",
        "requirements_files",
        "git_tag",
        "base_image_sha",
        "status_changed_at",
        "pinned_platform_version",
        "preview_deployment_tag",
        "profiling_mode",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_TAGS_FIELD_NUMBER: _ClassVar[int]
    CLOUD_BUILD_ID_FIELD_NUMBER: _ClassVar[int]
    TRIGGERED_BY_FIELD_NUMBER: _ClassVar[int]
    REQUIREMENTS_FILEPATH_FIELD_NUMBER: _ClassVar[int]
    DOCKERFILE_FILEPATH_FIELD_NUMBER: _ClassVar[int]
    RUNTIME_FIELD_NUMBER: _ClassVar[int]
    CHALKPY_VERSION_FIELD_NUMBER: _ClassVar[int]
    RAW_DEPENDENCY_HASH_FIELD_NUMBER: _ClassVar[int]
    FINAL_DEPENDENCY_HASH_FIELD_NUMBER: _ClassVar[int]
    IS_PREVIEW_DEPLOYMENT_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    GIT_COMMIT_FIELD_NUMBER: _ClassVar[int]
    GIT_PR_FIELD_NUMBER: _ClassVar[int]
    GIT_BRANCH_FIELD_NUMBER: _ClassVar[int]
    GIT_AUTHOR_EMAIL_FIELD_NUMBER: _ClassVar[int]
    BRANCH_FIELD_NUMBER: _ClassVar[int]
    PROJECT_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    REQUIREMENTS_FILES_FIELD_NUMBER: _ClassVar[int]
    GIT_TAG_FIELD_NUMBER: _ClassVar[int]
    BASE_IMAGE_SHA_FIELD_NUMBER: _ClassVar[int]
    STATUS_CHANGED_AT_FIELD_NUMBER: _ClassVar[int]
    PINNED_PLATFORM_VERSION_FIELD_NUMBER: _ClassVar[int]
    PREVIEW_DEPLOYMENT_TAG_FIELD_NUMBER: _ClassVar[int]
    PROFILING_MODE_FIELD_NUMBER: _ClassVar[int]
    id: str
    environment_id: str
    status: DeploymentStatus
    deployment_tags: _containers.RepeatedScalarFieldContainer[str]
    cloud_build_id: str
    triggered_by: str
    requirements_filepath: str
    dockerfile_filepath: str
    runtime: str
    chalkpy_version: str
    raw_dependency_hash: str
    final_dependency_hash: str
    is_preview_deployment: bool
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    git_commit: str
    git_pr: str
    git_branch: str
    git_author_email: str
    branch: str
    project_settings: str
    requirements_files: str
    git_tag: str
    base_image_sha: str
    status_changed_at: _timestamp_pb2.Timestamp
    pinned_platform_version: str
    preview_deployment_tag: str
    profiling_mode: DeploymentProfilingMode
    def __init__(
        self,
        id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        status: _Optional[_Union[DeploymentStatus, str]] = ...,
        deployment_tags: _Optional[_Iterable[str]] = ...,
        cloud_build_id: _Optional[str] = ...,
        triggered_by: _Optional[str] = ...,
        requirements_filepath: _Optional[str] = ...,
        dockerfile_filepath: _Optional[str] = ...,
        runtime: _Optional[str] = ...,
        chalkpy_version: _Optional[str] = ...,
        raw_dependency_hash: _Optional[str] = ...,
        final_dependency_hash: _Optional[str] = ...,
        is_preview_deployment: bool = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        git_commit: _Optional[str] = ...,
        git_pr: _Optional[str] = ...,
        git_branch: _Optional[str] = ...,
        git_author_email: _Optional[str] = ...,
        branch: _Optional[str] = ...,
        project_settings: _Optional[str] = ...,
        requirements_files: _Optional[str] = ...,
        git_tag: _Optional[str] = ...,
        base_image_sha: _Optional[str] = ...,
        status_changed_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        pinned_platform_version: _Optional[str] = ...,
        preview_deployment_tag: _Optional[str] = ...,
        profiling_mode: _Optional[_Union[DeploymentProfilingMode, str]] = ...,
    ) -> None: ...
